import stormpy
import stormpy.pomdp
import payntbind

import paynt.family.family
import paynt.quotient.quotient
import paynt.quotient.fsc

import math
import re
import collections

import logging
logger = logging.getLogger(__name__)


class PomdpQuotient(paynt.quotient.quotient.Quotient):

    # implicit size for POMDP unfolding
    initial_memory_size = 1
    # if True, posterior-aware unfolding will be applied
    posterior_aware = False

    def __init__(self, pomdp, specification, decpomdp_manager=None):
        super().__init__(specification=specification)

        # unfolded POMDP
        self.quotient_mdp = None
        self.family = None
        self.coloring = None
        # to each hole-option pair a list of actions colored by this combination (reverse coloring)
        self.hole_option_to_actions = None

        # attributes associated with a (folded) POMDP

        # default POMDP model
        self.pomdp = None
        # a (simplified) label for each observation
        self.observation_labels = None
        # number of actions available at each observation
        self.actions_at_observation = None
        # action labels corresponding to ^
        self.action_labels_at_observation = None
        # for each observation, number of states associated with it
        self.observation_states = None

        # attributes associated with an unfolded quotient MDP

        # number of memory states allocated to each observation
        self.observation_memory_size = None
        # Storm POMDP manager
        self.pomdp_manager = None

        # for each observation, a list of action holes
        self.observation_action_holes = None
        # for each observation, a list of memory holes
        self.observation_memory_holes = None
        # for each hole, an indication whether this is an action or a memory hole
        self.is_action_hole = None

        # construct the quotient POMDP
        self.pomdp = stormpy.pomdp.make_canonic(pomdp)
        # ^ this also asserts that states with the same observation have the
        # same number and the same order of available actions

        logger.info(f"constructed {'exact' if self.pomdp.is_exact else ''} POMDP having {self.observations} observations.")

        state_obs = self.pomdp.observations.copy()

        # extract observation labels
        if self.pomdp.has_observation_valuations():
            ov = self.pomdp.observation_valuations
            self.observation_labels = [ov.get_string(obs) for obs in range(self.observations)]
        else:
            if decpomdp_manager is None:
                self.observation_labels = list(range(self.observations))
                self.observation_labels = [str(label) for label in self.observation_labels]
            else:
                # map each 'joint' observation to the agent's observation and use the corresponding label
                self.observation_labels = []
                for obs in range(self.observations):
                    agent_obs = decpomdp_manager.joint_observations[obs][0]
                    agent_obs_label = decpomdp_manager.agent_observation_labels[0][agent_obs]
                    self.observation_labels.append(agent_obs_label)

        # compute actions available at each observation
        self.actions_at_observation = [0] * self.observations
        for state in range(self.pomdp.nr_states):
            obs = state_obs[state]
            if self.actions_at_observation[obs] != 0:
                continue
            self.actions_at_observation[obs] = self.pomdp.get_nr_available_actions(state)

        # collect labels of actions available at each observation
        self.action_labels_at_observation = [[] for obs in range(self.observations)]
        for state in range(self.pomdp.nr_states):
            obs = state_obs[state]
            if self.action_labels_at_observation[obs] != []:
                continue
            for offset in range(self.actions_at_observation[obs]):
                choice = self.pomdp.get_choice_index(state,offset)
                labels = self.pomdp.choice_labeling.get_labels_of_choice(choice)
                assert len(labels) <= 1, "expected at most 1 label"
                if len(labels) == 0:
                    label = self.EMPTY_LABEL
                else :
                    label = list(labels)[0]
                self.action_labels_at_observation[obs].append(label)
        for obs,labels in enumerate(self.action_labels_at_observation):
            if len(labels) == 0:
                logger.warning(f"WARNING: POMDP has no action for observation {obs}")

        # mark perfect observations
        self.observation_states = [0 for obs in range(self.observations)]
        for state in range(self.pomdp.nr_states):
            self.observation_states[state_obs[state]] += 1

        # initialize POMDP manager
        if self.pomdp.is_exact:
            if not PomdpQuotient.posterior_aware:
                self.pomdp_manager = payntbind.synthesis.ExactPomdpManager(self.pomdp)
            else:
                self.pomdp_manager = payntbind.synthesis.ExactPomdpManagerAposteriori(self.pomdp)
        else:
            if not PomdpQuotient.posterior_aware:
                self.pomdp_manager = payntbind.synthesis.PomdpManager(self.pomdp)
            else:
                self.pomdp_manager = payntbind.synthesis.PomdpManagerAposteriori(self.pomdp)

        # do initial unfolding
        self.set_imperfect_memory_size(PomdpQuotient.initial_memory_size)
        # self.set_global_memory_size(PomdpQuotient.initial_memory_size)
        self.current_memory_size = PomdpQuotient.initial_memory_size


    @property
    def observations(self):
        return self.pomdp.nr_observations

    def create_hole_name(self, obs, mem, is_action_hole):
        category = "A" if is_action_hole else "M"
        obs_label = self.observation_labels[obs]
        return "{}({},{})".format(category,obs_label,mem)

    def create_hole_name_aposteriori(self, is_action_hole, mem, prior, posterior=None):
        category = "A" if is_action_hole else "M"
        prior_label = self.observation_labels[prior]
        if posterior is None:
            return "{}({},{})".format(category,mem,prior_label)
        else:
            posterior_label = self.observation_labels[posterior]
            return "{}({},{},{})".format(category,mem,prior_label,posterior_label)


    def decode_hole_name(self, name):
        result = re.search(r"([A|M])\((.*?),(\d+)\)", name)
        is_action_hole = result.group(1) == "A"
        observation_label = result.group(2)
        memory = int(result.group(3))

        observation = None
        for obs in range(self.observations):
            if observation_label == self.observation_labels[obs]:
                observation = obs
                break
        return (is_action_hole, observation, memory)

    def set_manager_memory_vector(self):
        for obs in range(self.observations):
            mem = self.observation_memory_size[obs]
            self.pomdp_manager.set_observation_memory_size(obs,mem)

    def set_global_memory_size(self, memory_size):
        self.observation_memory_size = [memory_size] * self.observations
        self.set_manager_memory_vector()
        self.unfold_memory()

    def set_imperfect_memory_size(self, memory_size):
        ''' Set given memory size only to imperfect observations. '''
        self.observation_memory_size = [
            memory_size if self.observation_states[obs]>1 else 1
            for obs in range(self.observations)
        ]
        self.set_manager_memory_vector()
        self.unfold_memory()

    def increase_memory_size(self, obs):
        self.observation_memory_size[obs] += 1
        self.set_manager_memory_vector()
        self.unfold_memory()

    def set_memory_from_dict(self, obs_memory_dict):
        memory_list = []

        for obs in range(self.observations):
            memory_list.append(obs_memory_dict[obs])

        self.observation_memory_size = memory_list
        self.set_manager_memory_vector()
        self.unfold_memory()

    def set_memory_from_result_new(self, obs_memory_dict, obs_memory_dict_cutoff, memory_limit):
        memory_list = []

        for obs in range(self.observations):
            #memory = self.observation_memory_size[obs]
            if self.observation_states[obs] <= 1:
                memory = 1
            elif obs in obs_memory_dict.keys():
                memory = max(obs_memory_dict[obs], self.observation_memory_size[obs]+1)
            elif obs in obs_memory_dict_cutoff.keys():
                memory = obs_memory_dict_cutoff[obs]
            else:
                memory = memory_limit

            memory_list.append(memory)

        self.observation_memory_size = memory_list
        self.set_manager_memory_vector()
        self.unfold_memory()


    def create_coloring(self):
        if PomdpQuotient.posterior_aware:
            return self.create_coloring_aposteriori()

        # create holes
        family = paynt.family.family.Family()
        self.observation_action_holes = []
        self.observation_memory_holes = []
        self.is_action_hole = []

        for obs in range(self.observations):

            # action holes
            hole_indices = []
            num_actions = self.actions_at_observation[obs]
            if num_actions > 1:
                option_labels = self.action_labels_at_observation[obs]
                for mem in range(self.observation_memory_size[obs]):
                    hole_indices.append(family.num_holes)
                    name = self.create_hole_name(obs,mem,True)
                    family.add_hole(name,option_labels)
                    self.is_action_hole.append(True)
            self.observation_action_holes.append(hole_indices)

            # memory holes
            hole_indices = []
            num_updates = self.pomdp_manager.max_successor_memory_size[obs]
            if num_updates > 1:
                option_labels = [str(x) for x in range(num_updates)]
                for mem in range(self.observation_memory_size[obs]):
                    name = self.create_hole_name(obs,mem,False)
                    hole_indices.append(family.num_holes)
                    family.add_hole(name,option_labels)
                    self.is_action_hole.append(False)
            self.observation_memory_holes.append(hole_indices)

        # create the coloring
        assert self.pomdp_manager.num_holes == family.num_holes
        num_holes = family.num_holes
        choice_action_hole = self.pomdp_manager.row_action_hole
        choice_memory_hole = self.pomdp_manager.row_memory_hole
        choice_action_option = self.pomdp_manager.row_action_option
        choice_memory_option = self.pomdp_manager.row_memory_option
        choice_to_hole_options = []
        for choice in range(self.quotient_mdp.nr_choices):
            hole_options = []
            hole = choice_action_hole[choice]
            if hole != num_holes:
                hole_options.append( (hole,choice_action_option[choice]) )
            hole = choice_memory_hole[choice]
            if hole != num_holes:
                hole_options.append( (hole,choice_memory_option[choice]) )
            choice_to_hole_options.append(hole_options)

        return family, choice_to_hole_options

    def create_coloring_aposteriori(self):
        # a posteriori unfolding
        choice_to_hole_options = self.pomdp_manager.coloring
        hole_num_options = self.pomdp_manager.hole_num_options
        action_holes = self.pomdp_manager.action_holes
        update_holes = self.pomdp_manager.update_holes

        holes = [None] * len(hole_num_options)

        # action holes
        for key,index in action_holes.items():
            num_options = hole_num_options[index]
            if num_options <= 1:
                continue
            mem,prior = key
            name = self.create_hole_name_aposteriori(True,mem,prior)
            option_labels = [str(labels) for labels in self.action_labels_at_observation[prior]]
            holes[index] = (name,option_labels)

        # update holes
        for key,index in update_holes.items():
            num_options = hole_num_options[index]
            if num_options <= 1:
                continue
            mem,prior,posterior = key
            name = self.create_hole_name_aposteriori(False,mem,prior,posterior)
            option_labels = [str(x) for x in range(num_options)]
            holes[index] = (name,option_labels)

        # filter out trivial holes
        family = paynt.family.family.Family()
        old_to_new_indices = [None] * len(holes)
        for index,name_labels in enumerate(holes):
            if name_labels is None:
                continue
            old_to_new_indices[index] = family.num_holes
            name,option_labels = name_labels
            family.add_hole(name,option_labels)

        choice_to_hole_options_new = []
        for hole_options in choice_to_hole_options:
            hole_options_new = [ (old_to_new_indices[hole],v) for hole,v in hole_options.items() if old_to_new_indices[hole] is not None ]
            choice_to_hole_options_new.append(hole_options_new)
        choice_to_hole_options = choice_to_hole_options_new

        # creating this list to make it work with Paynt-Storm integration
        self.observation_action_holes = [[] for obs in range(self.observations)]
        for key,index in action_holes.items():
            _,prior = key
            new_index = old_to_new_indices[index]
            if new_index is not None:
                self.observation_action_holes[prior].append(new_index)

        return family, choice_to_hole_options


    def unfold_memory(self):

        # reset attributes
        self.quotient_mdp = None
        self.coloring = None
        self.hole_option_to_actions = None

        self.observation_action_holes = None
        self.observation_memory_holes = None
        self.is_action_hole = None

        logger.debug("unfolding {}-FSC template into POMDP...".format(max(self.observation_memory_size)))
        self.quotient_mdp = self.pomdp_manager.construct_mdp()
        if self.quotient_mdp.is_exact:
            self.choice_destinations = payntbind.synthesis.computeChoiceDestinationsExact(self.quotient_mdp)
        else:
            self.choice_destinations = payntbind.synthesis.computeChoiceDestinations(self.quotient_mdp)
        logger.debug(f"constructed quotient MDP having {self.quotient_mdp.nr_states} states and {self.quotient_mdp.nr_choices} actions.")

        self.family, choice_to_hole_options = self.create_coloring()

        self.coloring = payntbind.synthesis.Coloring(self.family.family, self.quotient_mdp.nondeterministic_choice_indices, choice_to_hole_options)

        # to each hole-option pair a list of actions colored by this combination
        self.hole_option_to_actions = [[] for hole in range(self.family.num_holes)]
        for hole in range(self.family.num_holes):
            self.hole_option_to_actions[hole] = [[] for option in self.family.hole_options(hole)]
        for choice in range(self.quotient_mdp.nr_choices):
            for hole,option in choice_to_hole_options[choice]:
                self.hole_option_to_actions[hole][option].append(choice)




    def estimate_scheduler_difference(self, mdp, quotient_choice_map, inconsistent_assignments, choice_values, expected_visits):

        if PomdpQuotient.posterior_aware:
            return super().estimate_scheduler_difference(mdp,quotient_choice_map,inconsistent_assignments,choice_values,expected_visits)

        # note: the code below is optimized for posterior-unaware unfolding

        # create inverse quotient-choice-to-mdp-choice map
        # TODO optimize this for multiple properties
        quotient_to_restricted_action_map = [None] * self.quotient_mdp.nr_choices
        for choice in range(mdp.nr_choices):
            quotient_to_restricted_action_map[quotient_choice_map[choice]] = choice

        # map choices to their origin states
        choice_to_state = []
        tm = mdp.transition_matrix
        for state in range(mdp.nr_states):
            for choice in tm.get_rows_for_group(state):
                choice_to_state.append(state)

        # for each hole, compute its difference sum and a number of affected states
        inconsistent_differences = {}
        for hole_index,options in inconsistent_assignments.items():
            difference_sum = 0
            states_affected = 0
            edges_0 = self.hole_option_to_actions[hole_index][options[0]]
            for choice_index,_ in enumerate(edges_0):

                choice_0_global = edges_0[choice_index]
                choice_0 = quotient_to_restricted_action_map[choice_0_global]
                if choice_0 is None:
                    continue

                source_state = choice_to_state[choice_0]
                source_state_visits = expected_visits[source_state]

                # assert source_state_visits != 0
                if source_state_visits == 0:
                    continue

                state_values = []
                for option in options:

                    assert len(self.hole_option_to_actions[hole_index][option]) > choice_index
                    choice_global = self.hole_option_to_actions[hole_index][option][choice_index]
                    choice = quotient_to_restricted_action_map[choice_global]
                    choice_value = choice_values[choice]
                    state_values.append(choice_value)

                min_value = min(state_values)
                max_value = max(state_values)
                difference = (max_value - min_value) * source_state_visits
                assert not math.isnan(difference)
                difference_sum += difference
                states_affected += 1

            if states_affected == 0:
                hole_score = 0
            else:
                hole_score = difference_sum / states_affected
            inconsistent_differences[hole_index] = hole_score

        return inconsistent_differences




    def sift_actions_and_updates(self, obs, hole, options):
        actions = set()
        updates = set()
        num_updates = self.pomdp_manager.max_successor_memory_size[obs]
        for option in options:
            actions.add(option // num_updates)
            updates.add(option %  num_updates)
        return actions,updates

    def break_symmetry_uai(self, family, action_inconsistencies, memory_inconsistencies):

        # go through each observation of interest and break symmetry
        restricted_family = family.copy()
        for obs in range(self.observations):

            num_actions = self.actions_at_observation[obs]
            num_updates = self.pomdp_manager.max_successor_memory_size[obs]

            obs_holes = self.obs_to_holes[obs]
            num_holes = len(obs_holes)


            all_actions = [action for action in range(num_actions)]
            selected_actions = [all_actions.copy() for hole in obs_holes]

            all_updates = [update for update in range(num_updates)]
            selected_updates = [all_updates.copy() for hole in obs_holes]

            inconsistencies = list(action_inconsistencies[obs])
            num_inc = len(inconsistencies)
            if num_inc > 1:
                # action inconsistency: allocate inconsistent actions between holes
                ignored_actions = [action for action in all_actions if action not in inconsistencies]
                selected_actions = [ignored_actions.copy() for hole in obs_holes]
                for index in range(max(num_holes,num_inc)):
                    selected_actions[index % num_holes].append(inconsistencies[index % num_inc])
            else:
                inconsistencies = list(memory_inconsistencies[obs])
                num_inc = len(inconsistencies)
                if num_inc > 1:
                    # memory inconsistency: distribute inconsistent updates between holes
                    ignored_updates = [update for update in all_updates if update not in inconsistencies]
                    selected_updates = [ignored_updates.copy() for hole in obs_holes]
                    for index in range(max(num_holes,num_inc)):
                        selected_updates[index % num_holes].append(inconsistencies[index % num_inc])

            # create options for each hole
            for index in range(num_holes):
                hole = obs_holes[index]
                actions = selected_actions[index]
                updates = selected_updates[index]
                options = []
                for action in actions:
                    for update in updates:
                        options.append(action * num_updates + update)
                restricted_family.hole_set_options(hole,options)

        # print(restricted_family)
        logger.debug("Symmetry breaking: reduced design space from {} to {}".format(family.size, restricted_family.size))

        return restricted_family



    def export_result(self, dtmc, mc_result):
        self.export_pomdp()
        self.export_optimal_dtmc(dtmc)
        self.export_policy(dtmc, mc_result)


    def export_pomdp(self):
        pomdp_path = "pomdp.drn"
        logger.info("Exporting POMDP to {}".format(pomdp_path))
        stormpy.export_to_drn(self.pomdp, pomdp_path)


    def export_optimal_dtmc(self, dtmc):

        # label states with a pomdp_state:memory_node pair
        # label choices with a pomdp_choice:memory_update pair
        state_labeling = dtmc.model.labeling
        choice_labeling = stormpy.storage.ChoiceLabeling(dtmc.model.nr_choices)
        for state in range(dtmc.states):
            mdp_state = dtmc.quotient_state_map[state]
            mdp_choice = dtmc.quotient_choice_map[state]

            pomdp_state = self.pomdp_manager.state_prototype[mdp_state]
            memory_node = self.pomdp_manager.state_memory[mdp_state]
            state_label = "{}:{}".format(pomdp_state,memory_node)
            if not state_labeling.contains_label(state_label):
                state_labeling.add_label(state_label)
            state_labeling.add_label_to_state(state_label,state)

            pomdp_action_index = self.pomdp_manager.row_action_option[mdp_choice]
            pomdp_choice = self.pomdp.get_choice_index(pomdp_state, pomdp_action_index)
            memory_update = self.pomdp_manager.row_memory_option[mdp_choice]
            choice_label = "{}:{}".format(pomdp_choice,memory_update)
            if not choice_labeling.contains_label(choice_label):
                choice_labeling.add_label(choice_label)
            # state and choices indices coincide for DTMCs
            choice_labeling.add_label_to_choice(choice_label,state)

        # add choice labeling to the model
        m = dtmc.model
        components = stormpy.storage.SparseModelComponents(m.transition_matrix,m.labeling,m.reward_models)
        components.choice_labeling = choice_labeling
        dtmc.model = stormpy.storage.SparseDtmc(components)

        # export DTMC
        dtmc_path = "dtmc.drn"
        logger.info("Exporting optimal DTMC to {}".format(dtmc_path))
        stormpy.export_to_drn(dtmc.model, dtmc_path)


    def collect_policy(self, dtmc, mc_result):
        # assuming single optimizing property
        assert self.specification.num_properties == 1 and self.specification.has_optimality
        dtmc_state_value = mc_result.optimality_result.result.get_values()

        # map states of the DTMC to their POMDP counterparts
        # label states with the value achieved in the state
        # group results by observation
        policy = []
        for obs in range(self.observations):
            mem_size = self.observation_memory_size[obs]
            mem_info = [ {} for _ in range(mem_size) ]
            policy.append(mem_info)

        for dtmc_state in range(dtmc.states):
            value = dtmc_state_value[dtmc_state]
            mdp_state = dtmc.quotient_state_map[dtmc_state]
            # mdp_choice = dtmc.quotient_choice_map[dtmc_state]

            pomdp_state = self.pomdp_manager.state_prototype[mdp_state]
            memory_node = self.pomdp_manager.state_memory[mdp_state]
            observation = self.pomdp.get_observation(pomdp_state)

            policy[observation][memory_node][pomdp_state] = value

        return policy

    def export_policy(self, dtmc, mc_result):

        policy = self.collect_policy(dtmc, mc_result)

        # use JSON as output format
        obs_info = []
        for obs in range(self.observations):
            policies = []
            for mem in range(self.pomdp_manager.observation_memory_size[obs]):
                if len(policy[obs][mem]) == 0:
                    continue
                state_values = [ {state:value} for state,value in policy[obs][mem].items() ]

                sub_policy = {}
                sub_policy["memory_node"] = mem
                sub_policy["state_values"] = state_values

                policies.append( sub_policy )
            obs_info.append(policies)

        # export JSON
        import json
        output_json = json.dumps(obs_info, indent=4)
        # print(output_json)
        scheduler_path = "scheduler.json"
        logger.info("Exporting optimal scheduler to {}".format(scheduler_path))
        with open(scheduler_path, 'w') as f:
            print(output_json, file=f)

    def extract_policy(self, assignment):
        dtmc = self.build_assignment(assignment)
        mc_result = dtmc.check_specification(self.specification)
        policy = self.collect_policy(dtmc, mc_result)
        return policy


    def policy_size(self, assignment):
        '''
        Compute how many natural numbers are needed to encode the mu-FSC under
        the current memory model mu.
        '''

        # going through the induced DTMC, too lazy to parse hole names
        dtmc = self.build_assignment(assignment)

        # size of action function gamma:
        #   for each memory node, a list of prior-action pairs
        size_gamma = sum(self.observation_memory_size) # explicit
        # size_gamma = sum([len(x) for x in prior_observations]) * 2 # sparse

        if not self.posterior_aware:
            # size of update function delta of a posterior-unaware FSC:
            #   for each memory node, a list of prior-update pairs
            size_delta = sum(self.observation_memory_size) # explicit
            return size_gamma + size_delta

        # posterior-aware update selection
        # for each memory node and for each prior, collect a set of possible posteriors
        max_mem = max(self.observation_memory_size)
        memory_prior_posteriors = [[set() for _ in range(self.observations)] for _ in range(max_mem)]
        for state in range(dtmc.states):
            mdp_state = dtmc.quotient_state_map[state]

            # get prior
            pomdp_state = self.pomdp_manager.state_prototype[mdp_state]
            memory_node = self.pomdp_manager.state_memory[mdp_state]
            prior = self.pomdp.get_observation(pomdp_state)
            # prior_observations[memory_node].add(prior)

            # get posterior observations
            for entry in dtmc.model.transition_matrix.get_row(state):
                successor = entry.column
                mdp_successor = dtmc.quotient_state_map[successor]
                pomdp_successor = self.pomdp_manager.state_prototype[mdp_successor]
                posterior = self.pomdp.get_observation(pomdp_successor)
                memory_prior_posteriors[memory_node][prior].add(posterior)

        # size of update function delta of a posterior-aware FSC:
        #   for each memory node and for each possible prior, a list of posterior-action pairs
        #   assuming sparse representation (not including delimeters)
        size_delta = 0
        for n_prior_posteriors in memory_prior_posteriors:
            for n_z_posteriors in n_prior_posteriors:
                size_delta += 2 * len(n_z_posteriors)

        return size_gamma + size_delta


    def get_family_pomdp(self, mdp):
        '''
        Constructs POMDP from the quotient MDP. Used for computing POMDP abstraction bounds.
        '''
        no_obs = self.pomdp.nr_observations
        tm = mdp.model.transition_matrix
        components = stormpy.storage.SparseModelComponents(tm, mdp.model.labeling, mdp.model.reward_models)

        full_observ_list = []
        for state in range(self.pomdp.nr_states):
            obs = self.pomdp.get_observation(state)
            for mem in range(self.observation_memory_size[obs]):
                full_observ_list.append(obs + mem * no_obs)

        choice_labeling = stormpy.storage.ChoiceLabeling(mdp.model.nr_choices)

        # assign observations to states
        observ_list = []
        choice_labels = []
        for state in range(mdp.model.nr_states):
            original_state = mdp.quotient_state_map[state]
            observ_list.append(full_observ_list[original_state])
            actions = [action for action in range(mdp.model.get_nr_available_actions(state))]
            choice_labels.append(actions)

        # construct labeling
        labels_list = [item for sublists in choice_labels for item in sublists]
        labels = list(set(labels_list))
        for label in labels:
            choice_labeling.add_label(str(label))
        for choice in range(mdp.model.nr_choices):
            choice_labeling.add_label_to_choice(str(labels_list[choice]), choice)

        components.choice_labeling = choice_labeling
        components.observability_classes = observ_list

        pomdp = stormpy.storage.SparsePomdp(components)
        pomdp = stormpy.pomdp.make_canonic(pomdp)

        return pomdp


    def assignment_to_fsc(self, assignment):
        assert assignment.size == 1, "expected family of size 1"
        num_nodes = max(self.observation_memory_size)
        fsc = paynt.quotient.fsc.FSC(num_nodes, self.observations, is_deterministic=True)
        fsc.observation_labels = self.observation_labels

        # collect action labels
        action_labels = set()
        for labels in self.action_labels_at_observation:
            action_labels.update(labels)
        action_labels = list(action_labels)
        fsc.action_labels = action_labels

        # map observations to unique indices of available actions
        action_label_indices = {label:index for index,label in enumerate(action_labels)}
        observation_to_actions = [[] for obs in range(self.observations)]
        for obs,action_labels in enumerate(self.action_labels_at_observation):
            observation_to_actions[obs] = [action_label_indices[label] for label in action_labels]

        fsc.fill_trivial_actions(observation_to_actions)
        fsc.fill_zero_updates()

        # convert hole assignment to FSC
        for obs,holes in enumerate(self.observation_action_holes):
            for node,hole in enumerate(holes):
                option = assignment.hole_options(hole)[0]
                action_label = self.action_labels_at_observation[obs][option]
                action = action_label_indices[action_label]
                fsc.action_function[node][obs] = action
        for obs,holes in enumerate(self.observation_memory_holes):
            for node,hole in enumerate(holes):
                option = assignment.hole_options(hole)[0]
                fsc.update_function[node][obs] = option

        fsc.fill_implicit_actions_and_updates()
        fsc.check(observation_to_actions)
        return fsc
    


    def get_induced_dtmc_from_fsc(self, fsc):
        # TODO maybe make this into payntbind function if it's slow
        if fsc.is_deterministic:
            fsc_copy = fsc.copy()
            fsc_copy.make_stochastic()
        else:
            fsc_copy = fsc
        action_function = fsc_copy.action_function
        update_function = fsc_copy.update_function
        action_labels = fsc_copy.action_labels

        # compute the state space for the induced dtmc
        dtmc_states_map = {}
        state_queue = [(self.pomdp.initial_states[0],0)]
        dtmc_states_map[len(dtmc_states_map)] = (self.pomdp.initial_states[0],0)

        while state_queue:
            current_state_memory_pair = state_queue.pop()

            # compute the successor states
            current_obs = self.pomdp.observations[current_state_memory_pair[0]]
            selected_actions = action_function[current_state_memory_pair[1]][current_obs]
            if selected_actions is None:
                continue
            selected_updates = update_function[current_state_memory_pair[1]][current_obs]
            if selected_updates is None:
                continue

            for selected_action in selected_actions.keys():
                for selected_update in selected_updates.keys():
                    selected_action_label = action_labels[selected_action]

                    choice_offset_for_selected_label = self.action_labels_at_observation[current_obs].index(selected_action_label)
                    choice_index = self.pomdp.get_choice_index(current_state_memory_pair[0], choice_offset_for_selected_label)

                    for entry in self.pomdp.transition_matrix.get_row(choice_index):
                        next_state = entry.column
                        next_state_memory_pair = (next_state,selected_update)
                        if next_state_memory_pair not in dtmc_states_map.values():
                            state_queue.append(next_state_memory_pair)
                            dtmc_states_map[len(dtmc_states_map)] = next_state_memory_pair


        # construct the transition matrix
        num_dtmc_states = len(dtmc_states_map)
        dtmc_tm_builder = stormpy.SparseMatrixBuilder(num_dtmc_states, num_dtmc_states, force_dimensions=True)
        state_action_rewards = {name:[] for name in self.pomdp.reward_models.keys()}
        for dtmc_state, current_state_memory_pair in dtmc_states_map.items():
            current_obs = self.pomdp.observations[current_state_memory_pair[0]]
            selected_actions = action_function[current_state_memory_pair[1]][current_obs]
            if selected_actions is None:
                for reward_name in self.pomdp.reward_models.keys():
                    state_action_rewards[reward_name].append(0)
                continue
            selected_updates = update_function[current_state_memory_pair[1]][current_obs]
            if selected_updates is None:
                for reward_name in self.pomdp.reward_models.keys():
                    state_action_rewards[reward_name].append(0)
                continue

            next_state_prob_map = {state:0 for state in dtmc_states_map.keys()}

            current_reward = {name:0 for name in self.pomdp.reward_models.keys()}

            for selected_action, action_prob in selected_actions.items():
                selected_action_label = action_labels[selected_action]
                choice_offset_for_selected_label = self.action_labels_at_observation[current_obs].index(selected_action_label)
                choice_index = self.pomdp.get_choice_index(current_state_memory_pair[0], choice_offset_for_selected_label)

                for reward_name, reward_model in self.pomdp.reward_models.items():
                    current_reward[reward_name] += reward_model.state_action_rewards[choice_index]*action_prob

                for selected_update, update_prob in selected_updates.items():
                    

                    for entry in self.pomdp.transition_matrix.get_row(choice_index):
                        next_state = entry.column
                        next_state_memory_pair = (next_state,selected_update)
                        next_state_index = [index for index,state in dtmc_states_map.items() if state == next_state_memory_pair]
                        assert len(next_state_index) == 1, "expected unique state for given state memory pair"
                        next_state_index = next_state_index[0]
                        next_state_prob_map[next_state_index] += entry.value()*action_prob*update_prob

            for reward_name in self.pomdp.reward_models.keys():
                state_action_rewards[reward_name].append(current_reward[reward_name])

                

            for next_state_index, next_state_prob in next_state_prob_map.items():
                dtmc_tm_builder.add_next_value(dtmc_state, next_state_index, next_state_prob)

        dtmc_tm = dtmc_tm_builder.build()

        # construct the labeling
        dtmc_labeling = stormpy.storage.StateLabeling(num_dtmc_states)
        for label in self.pomdp.labeling.get_labels():
            dtmc_labeling.add_label(label)
        for dtmc_state, current_state_memory_pair in dtmc_states_map.items():
            for label in self.pomdp.labeling.get_labels_of_state(current_state_memory_pair[0]):
                if label == "init" and current_state_memory_pair != (self.pomdp.initial_states[0],0): # only (0,0) state is initital
                    continue
                dtmc_labeling.add_label_to_state(label, dtmc_state)

        # construct the reward structure
        dtmc_reward_models = {}
        for reward_name in self.pomdp.reward_models.keys():
            assert reward_model.has_state_action_rewards == True, "currently this implementation expects state action rewards"
            dtmc_reward_models[reward_name] = stormpy.SparseRewardModel(optional_state_action_reward_vector=state_action_rewards[reward_name])

        components = stormpy.SparseModelComponents(transition_matrix=dtmc_tm, state_labeling=dtmc_labeling, reward_models=dtmc_reward_models)
        induced_dtmc = stormpy.storage.SparseDtmc(components)

        return induced_dtmc


    def compute_qvalues(self, assignment):
        '''
        Given an MDP obtained after applying an FSC to a POMDP, compute for each state s, (reachable) memory node n
        the Q-value Q(s,n).
        :param assignment hole assignment encoding an FSC; it is assumed the assignment is the one obtained
            for the current unfolding
        :note Q(s,n) may be None if (s,n) exists in the unfolded POMDP but is not reachable in the induced DTMC
        '''
        # model check
        submdp = self.build_assignment(assignment)
        prop = self.get_property()
        result = submdp.model_check_property(prop)
        state_submdp_to_value = result.result.get_values()

        # map states of a sub-MDP to the states of the quotient MDP to the state-memory pairs of the POMDPxFSC
        state_memory_value = collections.defaultdict(lambda: None)
        for submdp_state,value in enumerate(state_submdp_to_value):
            mdp_state = submdp.quotient_state_map[submdp_state]
            pomdp_state = self.pomdp_manager.state_prototype[mdp_state]
            memory_node = self.pomdp_manager.state_memory[mdp_state]
            state_memory_value[ (pomdp_state,memory_node) ] = value

        # make this mapping total
        memory_size = 1 + max([memory for state,memory in state_memory_value.keys()])
        state_memory_value_total = [[None for memory in range(memory_size)] for state in range(self.pomdp.nr_states)]
        for state in range(self.pomdp.nr_states):
            for memory in range(memory_size):
                value = state_memory_value[(state,memory)]
                if value is None:
                    obs = self.pomdp.observations[state]
                    if memory < self.observation_memory_size[obs]:
                        # case 1: (s,n) exists but is not reachable in the induced DTMC
                        value = None
                    else:
                        # case 2: (s,n) does not exist because n memory was not allocated for s
                        # i.e. (s,n) has the same value as (s,0)
                        value = state_memory_value[(state,0)]
                state_memory_value_total[state][memory] = value

        return state_memory_value_total


    def next_belief(self, belief, action_label, next_obs):
        any_belief_state = list(belief.keys())[0]
        obs = self.pomdp.observations[any_belief_state]
        action = self.action_labels_at_observation[obs].index(action_label)
        new_belief = collections.defaultdict(float)
        ndi = self.pomdp.nondeterministic_choice_indices.copy()
        for state,state_prob in belief.items():
            choice = self.pomdp.get_choice_index(state,action)
            for entry in self.pomdp.transition_matrix.get_row(choice):
                next_state = entry.column
                if self.pomdp.observations[next_state] == next_obs:
                    new_belief[next_state] += state_prob * entry.value()
        prob_sum = sum(new_belief.values())
        new_belief = {state:prob/prob_sum for state,prob in new_belief.items()}
        return new_belief



import paynt.quotient.posmg
import paynt.synthesizer.synthesizer
import paynt.quotient.pomdp
import paynt.verification.property_result

import logging
logger = logging.getLogger(__name__)

class SynthesizerAR(paynt.synthesizer.synthesizer.Synthesizer):

    @property
    def method_name(self):
        return "AR"

    def check_specification(self, family):
        ''' Check specification for mdp or smg based on self.quotient '''
        mdp = family.mdp

        if isinstance(self.quotient, paynt.quotient.posmg.PosmgQuotient):
            model = self.quotient.create_smg_from_mdp(mdp)
        else:
            model = mdp

        # check constraints
        admissible_assignment = None
        spec = self.quotient.specification
        if family.constraint_indices is None:
            family.constraint_indices = spec.all_constraint_indices()
        results = [None for _ in spec.constraints]
        for index in family.constraint_indices:
            constraint = spec.constraints[index]
            result = paynt.verification.property_result.MdpPropertyResult(constraint)
            results[index] = result

            # check primary direction
            result.primary = model.model_check_property(constraint)
            if result.primary.sat is False:
                result.sat = False
                break

            # check if the primary scheduler is consistent
            result.primary_selection,consistent = self.quotient.scheduler_is_consistent(mdp, constraint, result.primary.result)
            if consistent:
                assignment = family.assume_options_copy(result.primary_selection)
                dtmc = self.quotient.build_assignment(assignment)
                res = dtmc.check_specification(self.quotient.specification)
                if res.accepting_dtmc(self.quotient.specification):
                    result.sat = True
                    admissible_assignment = assignment

            # primary direction is SAT: check secondary direction to see whether all SAT
            result.secondary = model.model_check_property(constraint, alt=True)
            if mdp.is_deterministic and result.primary.value != result.secondary.value:
                logger.warning("WARNING: model is deterministic but min<max")
            if result.secondary.sat:
                result.sat = True
                continue

        spec_result = paynt.verification.property_result.MdpSpecificationResult()
        spec_result.constraints_result = paynt.verification.property_result.ConstraintsResult(results)

        # check optimality
        if spec.has_optimality and not spec_result.constraints_result.sat is False:
            opt = spec.optimality
            result = paynt.verification.property_result.MdpOptimalityResult(opt)

            # check primary direction
            result.primary = model.model_check_property(opt)
            if not result.primary.improves_optimum:
                # OPT <= LB
                result.can_improve = False
            else:
                # LB < OPT, check if LB is tight
                result.primary_selection,consistent = self.quotient.scheduler_is_consistent(mdp, opt, result.primary.result)
                result.can_improve = True
                if consistent:
                    # LB < OPT and it's tight, double-check the constraints and the value on the DTMC
                    result.can_improve = False
                    assignment = family.assume_options_copy(result.primary_selection)
                    dtmc = self.quotient.build_assignment(assignment)
                    res = dtmc.check_specification(self.quotient.specification)
                    if res.constraints_result.sat and spec.optimality.improves_optimum(res.optimality_result.value):
                        result.improving_assignment = assignment
                        result.improving_value = res.optimality_result.value
            spec_result.optimality_result = result

        spec_result.evaluate(family, admissible_assignment)
        family.analysis_result = spec_result

    def verify_family(self, family):
        self.quotient.build(family)

        # TODO include iteration_game in iteration? is it necessary?
        if isinstance(self.quotient, paynt.quotient.posmg.PosmgQuotient):
            self.stat.iteration_game(family.mdp.states)
        else:
            self.stat.iteration(family.mdp)

        self.check_specification(family)

    def update_optimum(self, family):
        ia = family.analysis_result.improving_assignment
        if ia is None:
            return
        if not self.quotient.specification.has_optimality:
            self.best_assignment = ia
            return
        iv = family.analysis_result.improving_value
        if not self.quotient.specification.optimality.improves_optimum(iv):
            return
        self.quotient.specification.optimality.update_optimum(iv)
        self.best_assignment = ia
        self.best_assignment_value = iv
        # logger.info(f"value {round(iv,4)} achieved after {round(paynt.utils.timer.GlobalTimer.read(),2)} seconds")
        if isinstance(self.quotient, paynt.quotient.pomdp.PomdpQuotient):
            self.stat.new_fsc_found(family.analysis_result.improving_value, ia, self.quotient.policy_size(ia))

    def synthesize_one(self, family):
        families = [family]
        while families:
            if self.resource_limit_reached():
                break
            family = families.pop(-1)
            self.verify_family(family)
            self.update_optimum(family)
            if not self.quotient.specification.has_optimality and self.best_assignment is not None:
                break
            # break
            if family.analysis_result.can_improve is False:
                self.explore(family)
                continue
            # undecided
            subfamilies = self.quotient.split(family)
            families = families + subfamilies
        return self.best_assignment



import re
import pygraphviz as pgv

def parse_hole(name) -> object:
    assert re.match(r"[AM]\(\[.*\],\d\)",
                    name), "Cannot use restrict function, hole name doesn't match"
    hole = {}
    hole["type"] = "Memory" if name[0] == "M" else "Assignment"
    hole["memory"] = int(name[-2])
    hole["observation"] = int(name[5]) if re.match(
        r"[MA]\(\[o=\d\],\d\)", name) else 0

    return hole


class Graph:
    """A class for creating graphs from design space structures.
    """

    def __init__(self) -> None:
        """Initializes an instance of the Graph class.
        """
        self.graph = pgv.AGraph(strict=False, directed=True)

    def parse(self, family) -> None:
        """Parses the design space holes into the nodes dictionary:
                {start_node: {end_node1: [observation1],...}, end_node2: ...}

            family: list of holes
        """
        self.nodes = {}
        for hole in range(family.num_holes):
            tmp = parse_hole(family.hole_name(hole))
            tmp["next"] = list(family.hole_options(hole))

            for next in tmp["next"]:
                if tmp["type"] == "Assignment":
                    continue

                if tmp["memory"] in self.nodes.keys():
                    if next in self.nodes[tmp["memory"]].keys():
                        self.nodes[tmp["memory"]][next].append(
                            tmp["observation"])
                    else:
                        self.nodes[tmp["memory"]][next] = [tmp["observation"]]
                else:
                    self.nodes[tmp["memory"]] = {next: [tmp["observation"]]}

    def create_graph(self, show_labels=False) -> None:
        """Creates a graph from the parsed design space.
        """
        self.graph.clear()
        nodes = list(self.nodes.keys())
        self.graph.add_nodes_from(nodes, fontsize=12, width=0.5, height=0.5)

        for (start, ends) in self.nodes.items():
            for (end, labels) in ends.items():
                if show_labels:
                    self.graph.add_edge(start, end, label=",".join(
                        [str(l) for l in sorted(labels)]), fontsize=12)
                else:
                    self.graph.add_edge(start, end)
        self.graph.layout("circo")

    def print(self, family, file_name="out", show_labels=False, args="-Gsize=5! -Gratio=\"expand\" -Gnodesep=.5") -> None:
        """Prints a graph in .png format.

            family: list of Holes
            file_name: name of file where the graph will be saved without the extention (e.g. file_name=file => file.png)
            args: string with Graphvix arguments to specify the output format
        """
        self.parse(family)
        self.create_graph(show_labels=show_labels)
        self.graph.draw(file_name + ".png", format="png", args=args)

    def __str__(self) -> str:
        return self.graph.string()



import paynt.synthesizer.synthesizer

import logging
logger = logging.getLogger(__name__)

class SynthesizerOneByOne(paynt.synthesizer.synthesizer.Synthesizer):
    
    @property
    def method_name(self):
        return "1-by-1"

    def synthesize_one(self, family):
        
        for hole_combination in family.all_combinations():
            
            assignment = family.construct_assignment(hole_combination)
            dtmc = self.quotient.build_assignment(assignment)
            self.stat.iteration(dtmc)
            result = dtmc.check_specification(self.quotient.specification, short_evaluation=True)
            self.explore(assignment)

            accepting,improving_value = result.accepting_dtmc(self.quotient.specification)
            if accepting:
                self.best_assignment = assignment
            if improving_value is not None:
                self.quotient.specification.optimality.update_optimum(improving_value)
            if accepting and not self.quotient.specification.can_be_improved:
                return accepting_assignment

        return self.best_assignment

    def evaluate_all(self, family, prop, keep_value_only=False):

        if not keep_value_only:
            logger.debug("forcing keep_value_only=True for the one-by-one evaluation")
            keep_value_only = True

        evaluations = []
        for hole_combination in family.all_combinations():
            assignment = family.construct_assignment(hole_combination)
            model = self.quotient.build_assignment(assignment)
            self.stat.iteration(model)
            result = model.model_check_property(prop)
            if keep_value_only:
                evaluation = result.value
            else:
                policy = None
                if result.sat:
                    policy = self.quotient.scheduler_to_policy(result.result.scheduler, model)
                evaluation = paynt.synthesizer.synthesizer.FamilyEvaluation(assignment, result.value, result.sat, policy)
            evaluations.append(evaluation)
            self.explore(assignment)
        return evaluations

    def export_evaluation_result(self, evaluations, export_filename_base):
        import json
        family_to_evaluation_parsed = []
        for evaluation in evaluations:
            family = evaluation.family
            policy = evaluation.policy
            if policy is None:
                policy = "UNSAT"
            else:
                policy = self.quotient.policy_to_state_valuation_actions(policy)
            family_to_evaluation_parsed.append( (str(family),policy) )
        policies_string = json.dumps(family_to_evaluation_parsed, indent=2)
        policies_filename = export_filename_base + ".json"
        with open(policies_filename, 'w') as file:
            file.write(policies_string)
        logger.info(f"exported satisfied members and correponding policies to {policies_filename}")



import stormpy
import payntbind

from collections import defaultdict

import os

import logging
logger = logging.getLogger(__name__)


class DrnParser:

    COMMENT_PREFIX = '//'
    TYPE_PREFIX = '@type: '
    STATE_PREFIX = 'state '
    WHITESPACES = ' \t\n\v\f\r'

    @classmethod
    def parse_drn(cls, sketch_path, use_exact=False):
        # try to read a drn file and return POSMG or POMDP based on the type
        # ValueError if file is not dnr or the model is of unsupported type
        explicit_model = None
        try:
            type = DrnParser.decide_type_of_drn(sketch_path)
            if type == 'POSMG':
                if use_exact:
                    raise ValueError('Exact synthesis is not supported for POSMG models')
                pomdp_path = sketch_path + '.tmp'
                state_player_indications = DrnParser.pomdp_from_posmg(sketch_path, pomdp_path)
                pomdp = DrnParser.read_drn(pomdp_path)
                explicit_model = payntbind.synthesis.posmg_from_pomdp(pomdp, state_player_indications)
                os.remove(pomdp_path)
            else:
                explicit_model = DrnParser.read_drn(sketch_path, use_exact)
        except Exception as e:
            print(e)
            raise ValueError('Failed to read sketch file in a .drn format')
        return explicit_model

    @classmethod
    def decide_type_of_drn(cls, path: str) -> str:
        # decide type of model in drn file and return it as a string
        # path - path to drn file
        # return - string representation of type (e.g. 'POMDP')
        # ValueError if not valid drn file
        with open(path) as file:
            while True:
                line = file.readline()
                if line.isspace() or line.lstrip(cls.WHITESPACES).startswith(cls.COMMENT_PREFIX):
                    continue
                if line.startswith(cls.TYPE_PREFIX):
                    type = line.removeprefix(cls.TYPE_PREFIX).removesuffix('\n')
                    return type
                raise ValueError

    @classmethod
    def pomdp_from_posmg(cls, old_path: str, new_path) -> list:
        # Change type of model in drn file from posmg to pomdp and store it in a new file
        # old_path - path to drn file with posmg model
        # new_path - path to drn file to store pomdp model
        # return - list indicating which states (indices) belongs to which player (values)
        posmg_file = open(old_path, 'r')
        pomdp_file = open(new_path, 'w')
        state_player_indications = []

        for line in posmg_file:
            if line.startswith(cls.TYPE_PREFIX):
                line = line.replace('POSMG', 'POMDP')
            if line.startswith(cls.STATE_PREFIX):
                l_idx = line.index('<')
                r_idx = line.index('>')
                player_num = line[l_idx + 1:r_idx]
                state_player_indications.append(int(player_num))
                line = cls.str_remove_range(line, l_idx-1, r_idx) # -1 to remove the space before <

            line = line.replace(' []', '') # temporaty fix
            pomdp_file.write(line)

        posmg_file.close()
        pomdp_file.close()

        return state_player_indications

    @staticmethod
    def str_remove_range(string: str, start_idx: int, end_idx: int) -> str:
        # return string without characters in specified range
        # both indices will be removed
        assert 0 <= start_idx < len(string)
        assert 0 <= end_idx < len(string)
        assert start_idx <= end_idx

        return string[:start_idx] + string[end_idx+1:]

    @classmethod
    def read_drn(cls, sketch_path, use_exact=False):
        builder_options = stormpy.DirectEncodingParserOptions()
        builder_options.build_choice_labels = True
        if use_exact:
            return stormpy._build_sparse_exact_model_from_drn(sketch_path, builder_options)
        else:
            return stormpy.build_model_from_drn(sketch_path, builder_options)

    @staticmethod
    def parse_posmg_specification(properties_path):
        if not os.path.isfile(properties_path):
            raise ValueError(f"the properties file {properties_path} does not exist")
        logger.info(f"loading properties from {properties_path} ...")

        with open(properties_path) as file:
            line = file.readline()

            return stormpy.parse_properties(line)

    @classmethod
    def write_model_in_pomdp_solve_format(cls, pomdp, output_path, property_path):

        num_states = pomdp.nr_states
        num_choices = pomdp.nr_choices
        num_obs = pomdp.nr_observations
        max_num_choices = max([pomdp.get_nr_available_actions(state) for state in range(num_states)])


        desc = """\
# auto-generated from PRISM program

discount: 0.95
values: reward
states: {}
actions: {}
observations: {}

""".format(num_states,max_num_choices,num_obs)

        # initial state
        state_init = pomdp.initial_states[0]
        initial_distr = [1 if state == state_init else 0 for state in range(num_states)]
        initial_distr = [str(x) for x in initial_distr]
        initial_distr = ' '.join(initial_distr)
        desc += f"start:\n{initial_distr}\n\n"

        # transition matrix
        desc += "# transition matrix\n\n"

        tm = pomdp.transition_matrix
        for state in range(num_states):
            action_index = 0
            group_start = tm.get_row_group_start(state)
            group_end = tm.get_row_group_end(state)
            trivial_action = group_end == group_start + 1
            for row_index in range(group_start, group_end):
                for entry in tm.get_row(row_index):
                    action_index_str = action_index if not trivial_action else "*"
                    desc += f"T: {action_index_str} : {state} : {entry.column} {entry.value()}\n"
                action_index += 1

        # observations
        desc += "\n# observations\n\n"
        for state in range(num_states):
            desc += f"O: * : {state} : {pomdp.observations[state]} 1\n"

        # rewards
        if len(pomdp.reward_models) > 0:
            # rewards
            desc += "\n# rewards\n\n"

            # assuming a single reward model
            rewards = next(iter(pomdp.reward_models.values()))

            # convert rewards to state-based
            state_rewards = []
            if rewards.has_state_rewards:
                state_rewards = list(rewards.state_rewards)
            elif rewards.has_state_action_rewards:
                state_action_rewards = list(rewards.state_action_rewards)
                for state in range(num_states):
                    group_start = tm.get_row_group_start(state)
                    state_rewards.append(state_action_rewards[group_start])
            else:
                raise TypeError("unknown reward type")

            # print state-based rewards
            for state in range(num_states):
                rew = state_rewards[state]
                if rew != 0:
                    desc += f"R: * : {state} : * : * {rew}\n"

        # ready to print
        logger.info("Writing POMDP in pomdp-solve format to {} ...".format(output_path))
        with open(output_path, 'w') as f:
            f.write(desc)
        logger.info("Writing default discount property to {} ...".format(property_path))
        with open(property_path, 'w') as f:
            f.write('R{"rew0"}max=? [F "target"]')
        logger.info("Write OK, aborting ...")
        exit(0)



import paynt.synthesizer.synthesizer_ar
import paynt.quotient.mdp
import paynt.utils.timer

import stormpy
import payntbind

import os
import json

import logging
logger = logging.getLogger(__name__)

class SynthesizerDecisionTree(paynt.synthesizer.synthesizer_ar.SynthesizerAR):

    # tree depth
    tree_depth = 0
    # if set, all trees of size at most tree_depth will be enumerated
    tree_enumeration = False
    # path to a scheduler to be mapped to a decision tree
    scheduler_path = None

    def __init__(self, *args):
        super().__init__(*args)
        self.best_tree = None
        self.best_tree_value = None

    @property
    def method_name(self):
        return "AR (decision tree)"

    def verify_hole_selection(self, family, hole_selection):
        spec = self.quotient.specification
        assignment = family.assume_options_copy(hole_selection)
        dtmc = self.quotient.build_assignment(assignment)
        res = dtmc.check_specification(spec)
        if not res.constraints_result.sat:
            return
        if not spec.has_optimality:
            family.analysis_result.improving_assignment = assignment
            family.analysis_result.can_improve = False
            return
        assignment_value = res.optimality_result.value
        if spec.optimality.improves_optimum(assignment_value):
            # logger.info(f"harmonization achieved value {res.optimality_result.value}")
            self.num_harmonization_succeeded += 1
            family.analysis_result.improving_assignment = assignment
            family.analysis_result.improving_value = assignment_value
            family.analysis_result.can_improve = True
            self.update_optimum(family)


    def harmonize_inconsistent_scheduler(self, family):
        self.num_harmonizations += 1
        mdp = family.mdp
        result = family.analysis_result.undecided_result()
        hole_selection = result.primary_selection
        harmonizing_hole = [hole for hole,options in enumerate(hole_selection) if len(options)>1][0]
        selection_1 = hole_selection.copy(); selection_1[harmonizing_hole] = [selection_1[harmonizing_hole][0]]
        selection_2 = hole_selection.copy(); selection_2[harmonizing_hole] = [selection_2[harmonizing_hole][1]]
        for selection in [selection_1,selection_2]:
            self.verify_hole_selection(family,selection)


    def verify_family(self, family):
        self.num_families_considered += 1
        self.quotient.build(family)

        self.stat.iteration(family.mdp)
        if family.parent_info is not None:
            for choice in family.parent_info.scheduler_choices:
                if not family.selected_choices[choice]:
                    break
            else:
                # scheduler preserved in the sub-family
                self.num_schedulers_preserved += 1
                family.analysis_result = family.parent_info.analysis_result
                family.scheduler_choices = family.parent_info.scheduler_choices
                consistent,hole_selection = self.quotient.are_choices_consistent(family.scheduler_choices, family)
                assert not consistent
                family.analysis_result.optimality_result.primary_selection = hole_selection
                return

        self.num_families_model_checked += 1
        self.check_specification(family)
        if not family.analysis_result.can_improve:
            return
        self.harmonize_inconsistent_scheduler(family)


    def counters_reset(self):
        self.num_families_considered = 0
        self.num_families_skipped = 0
        self.num_families_model_checked = 0
        self.num_schedulers_preserved = 0
        self.num_harmonizations = 0
        self.num_harmonization_succeeded = 0

    def counters_print(self):
        logger.info(f"families considered: {self.num_families_considered}")
        logger.info(f"families skipped by construction: {self.num_families_skipped}")
        logger.info(f"families with schedulers preserved: {self.num_schedulers_preserved}")
        logger.info(f"families model checked: {self.num_families_model_checked}")
        logger.info(f"harmonizations attempted: {self.num_harmonizations}")
        logger.info(f"harmonizations succeeded: {self.num_harmonization_succeeded}")
        print()

    def export_decision_tree(self, decision_tree, export_filename_base):
        tree = decision_tree.to_graphviz()
        tree_filename = export_filename_base + ".dot"
        directory = os.path.dirname(tree_filename)
        if directory:
            os.makedirs(directory, exist_ok=True)
        with open(tree_filename, 'w') as file:
            file.write(tree.source)
        logger.info(f"exported decision tree to {tree_filename}")

        tree_visualization_filename = export_filename_base + ".png"
        tree.render(export_filename_base, format="png", cleanup=True) # using export_filename_base since graphviz appends .png by default
        logger.info(f"exported decision tree visualization to {tree_visualization_filename}")


    def synthesize_tree(self, depth:int):
        self.counters_reset()
        self.quotient.reset_tree(depth)
        self.best_assignment = self.best_assignment_value = None
        self.synthesize(keep_optimum=True)
        if self.best_assignment is not None:
            self.quotient.decision_tree.root.associate_assignment(self.best_assignment)
            self.best_tree = self.quotient.decision_tree
            self.best_tree_value = self.best_assignment_value
        self.best_assignment = self.best_assignment_value = None
        self.counters_print()

    def synthesize_tree_sequence(self, opt_result_value):
        self.best_tree = self.best_tree_value = None

        global_timeout = paynt.utils.timer.GlobalTimer.global_timer.time_limit_seconds
        if global_timeout is None: global_timeout = 900
        depth_timeout = global_timeout / 2 / SynthesizerDecisionTree.tree_depth
        for depth in range(SynthesizerDecisionTree.tree_depth+1):
            print()
            self.quotient.reset_tree(depth)
            best_assignment_old = self.best_assignment

            family = self.quotient.family
            self.explored = 0
            self.counters_reset()
            self.stat = paynt.synthesizer.statistic.Statistic(self)
            self.stat.start(family)
            timeout = depth_timeout if depth < SynthesizerDecisionTree.tree_depth else None
            self.synthesis_timer = paynt.utils.timer.Timer(timeout)
            self.synthesis_timer.start()
            families = [family]

            if self.best_tree is not None:
                subfamily = family.copy()
                self.quotient.decision_tree.root.apply_hint(subfamily,self.best_tree.root)
                families = [subfamily,family]

            for family in families:
                self.synthesize_one(family)
            self.stat.finished_synthesis()
            self.stat.print()
            self.synthesis_timer = None
            self.counters_print()

            new_assignment_synthesized = self.best_assignment != best_assignment_old
            if new_assignment_synthesized:
                logger.info("printing synthesized assignment below:")
                logger.info(self.best_assignment)

                if self.best_assignment is not None and self.best_assignment.size == 1:
                    dtmc = self.quotient.build_assignment(self.best_assignment)
                    result = dtmc.check_specification(self.quotient.specification)
                    logger.info(f"double-checking specification satisfiability: {result}")

                self.best_tree = self.quotient.decision_tree
                self.best_tree.root.associate_assignment(self.best_assignment)
                self.best_tree_value = self.best_assignment_value

                if abs( (self.best_assignment_value-opt_result_value)/opt_result_value ) < 1e-3:
                    break

            if self.resource_limit_reached():
                break

    def map_scheduler(self, scheduler_choices):
        self.counters_reset()
        for depth in range(SynthesizerDecisionTree.tree_depth+1):
            self.quotient.reset_tree(depth,enable_harmonization=False)
            family = self.quotient.family.copy()
            family.analysis_result = self.quotient.build_unsat_result()
            self.quotient.build(family)
            consistent,hole_selection = self.quotient.are_choices_consistent(scheduler_choices, family)
            if consistent:
                self.verify_hole_selection(family,hole_selection)
                if self.best_assignment is not None:
                    self.best_tree = self.quotient.decision_tree
                    self.best_tree.root.associate_assignment(self.best_assignment)
                    self.best_tree_value = self.best_assignment_value
                    break

            if self.resource_limit_reached():
                break

    def run(self, optimum_threshold=None):
        # self.quotient.reset_tree(SynthesizerDecisionTree.tree_depth,enable_harmonization=True)
        scheduler_choices = None
        if SynthesizerDecisionTree.scheduler_path is None:
            paynt_mdp = paynt.models.models.Mdp(self.quotient.quotient_mdp)
            mc_result = paynt_mdp.model_check_property(self.quotient.get_property())
        else:
            opt_result_value = None
            with open(SynthesizerDecisionTree.scheduler_path, 'r') as f:
                scheduler_json = json.load(f)
            scheduler_choices,scheduler_json_relevant = self.quotient.scheduler_json_to_choices(scheduler_json, discard_unreachable_states=True)

            # export transformed scheduler
            # import os
            # directory = os.path.dirname(SynthesizerDecisionTree.scheduler_path)
            # transformed_name = f"scheduler-reachable.storm.json"
            # scheduler_relevant_path = os.path.join(directory, transformed_name)
            # with open(scheduler_relevant_path, 'w') as f:
            #     json.dump(scheduler_json_relevant, f, indent=4)
            # logger.debug(f"stored transformed scheduler to {scheduler_relevant_path}")
            # exit()

            submdp = self.quotient.build_from_choice_mask(scheduler_choices)
            mc_result = submdp.model_check_property(self.quotient.get_property())
        opt_result_value = mc_result.value
        logger.info(f"the optimal scheduler has value: {opt_result_value}")

        if self.quotient.DONT_CARE_ACTION_LABEL in self.quotient.action_labels:
            random_choices = self.quotient.get_random_choices()
            submdp_random = self.quotient.build_from_choice_mask(random_choices)
            mc_result_random = submdp_random.model_check_property(self.quotient.get_property())
            random_result_value = mc_result_random.value
            logger.info(f"the random scheduler has value: {random_result_value}")
            # self.set_optimality_threshold(random_result_value)

        self.best_assignment = self.best_assignment_value = None
        self.best_tree = self.best_tree_value = None
        if scheduler_choices is not None:
            self.map_scheduler(scheduler_choices)
        else:
            if self.quotient.specification.has_optimality:
                epsilon = 1e-1
                mc_result_positive = opt_result_value > 0
                if self.quotient.specification.optimality.maximizing == mc_result_positive:
                    epsilon *= -1
                # optimum_threshold = opt_result_value * (1 + epsilon)
            self.set_optimality_threshold(optimum_threshold)

            if not SynthesizerDecisionTree.tree_enumeration:
                self.synthesize_tree(SynthesizerDecisionTree.tree_depth)
            else:
                self.synthesize_tree_sequence(opt_result_value)

        logger.info(f"the optimal scheduler has value: {opt_result_value}")
        if self.quotient.DONT_CARE_ACTION_LABEL in self.quotient.action_labels:
            logger.info(f"the random scheduler has value: {random_result_value}")
        if self.best_tree is None:
            logger.info("no admissible tree found")
        else:
            relevant_state_valuations = [self.quotient.relevant_state_valuations[state] for state in self.quotient.state_is_relevant_bv]
            self.best_tree.simplify(relevant_state_valuations)
            depth = self.best_tree.get_depth()
            num_nodes = len(self.best_tree.collect_nonterminals())
            logger.info(f"synthesized tree of depth {depth} with {num_nodes} decision nodes")
            if self.quotient.specification.has_optimality:
                logger.info(f"the synthesized tree has value {self.best_tree_value}")
            if self.quotient.DONT_CARE_ACTION_LABEL in self.quotient.action_labels:
                logger.info(f"the synthesized tree has relative value: {(self.best_tree_value-random_result_value)/(opt_result_value-random_result_value)}")
            logger.info(f"printing the synthesized tree below:")
            print(self.best_tree.to_string())
            # logger.info(f"printing the PRISM module below:")
            # print(self.best_tree.to_prism())

            if self.export_synthesis_filename_base is not None:
                self.export_decision_tree(self.best_tree, self.export_synthesis_filename_base)

        time_total = round(paynt.utils.timer.GlobalTimer.read(),2)
        logger.info(f"synthesis finished after {time_total} seconds")

        print()
        for name,time in self.quotient.coloring.getProfilingInfo():
            time_percent = round(time/time_total*100,1)
            print(f"{name} = {time} s ({time_percent} %)")

        return self.best_tree



import os
import stormpy
import stormpy.examples
import stormpy.examples.files

from paynt.examples import paynt_models_dir

stormpy_example_dir = stormpy.examples.files.testfile_dir


def get_stormpy_example_path(*paths):
    return os.path.join(stormpy_example_dir, *paths)

def get_sketch_paths(project_path, sketch_name="sketch.templ", props_name="sketch.props"):
    return os.path.join(paynt_models_dir, project_path, sketch_name), os.path.join(paynt_models_dir, project_path, props_name)



import click
import os


def construct_controller(stations):
    stations_str = f"\nconst int STATIONS = {stations};\n"
    head = "\nmodule controller\n"
    round_var = "\tround : [0..STATIONS] init 1;\n\n"
    sync = "\t[sync] round = 0 -> (round'=1);\n"
    steps = [f"\t[step{i}] round = {i} -> (round'={i + 1});" for i in range(1, stations)]
    step_p = f"\t[step{stations}] round = STATIONS -> (round'=0);\n"
    tail = "\nendmodule\n\n"
    return stations_str + head + round_var + sync + "\n".join(steps) + "\n" + step_p + tail


def construct_holes(dices):
    m0l = "hole int M0LFAIR in {" + ",".join(map(str, range(0, dices))) + "};\n"
    m0h = "hole int M0HFAIR in {" + ",".join(map(str, range(0, dices))) + "};\n"
    m1l = "hole int M1LFAIR in {" + ",".join(map(str, range(0, dices))) + "};\n"
    m1h = "hole int M1HFAIR in {" + ",".join(map(str, range(0, dices))) + "};\n"
    ms = "hole int MxxA in {0,1};\n" + "hole int MxxB in {0,1};\n" + "hole int MxxC in {0,1};\n"
    return m0l + m0h + m1l + m1h + ms


def construct_process(dices, stations):
    head = "\nmodule process1\n"
    x1 = "\n\t x1 : [0..1] init 0;\n"
    y1 = "\t y1 : [0..1] init 0;\n"
    m1 = "\t m1 : [0..1] init 0;\n"
    init = "\n\t[go] true -> 1: (x1'=1);\n\n"
    ms = ["M0LFAIR", "M0HFAIR", "M1LFAIR", "M1HFAIR"]
    guards_eq = [
        f"\t[step1]  (x1=x{stations}) & m1 = {m} & x1 = {x} & {ms[mx]}={i} -> " \
        f"p{i} : (y1'=0) & (m1'=MxxA) + 1-p{i} : (y1'=1) & (m1'=MxxB);"
        for mx, (m, x) in enumerate([(0, 0), (0, 1), (1, 0), (1, 1)]) for i in range(0, dices)
    ]
    quards_neq = [
        f"\t[step1] !(x1=x{stations}) & m1 = {m} & x1 = {x} -> (y1'=x{stations}) & (m1'=MxxC);"
        for m, x in [(0, 0), (0, 1), (1, 0), (1, 1)]
    ]
    sync = "\n\t[sync] true -> (x1' = y1) & (y1' = 0);\n\n"
    tail = "endmodule\n"
    return head + x1 + y1 + m1 + init + "\n".join(guards_eq) + "\n\n" + "\n".join(quards_neq) + "\n" + sync + tail


@click.command()
@click.option('--dices', help="The number of considered dices for each combination.", required=False, default=5)
@click.option('--stations', help="The number of considered stations.", required=False, default=5)
@click.option('--step', help="The step between probabilities of considered dices.", required=False, default=0.01)
def generate_herman_sketch(dices, stations, step):
    if stations % 2 == 0:
        print("The number of stations must be even!")

    p_step = f"\nconst double P_STEP={step};\n\n"
    num_tokens = "\nformula num_tokens = " + \
                 "".join([f"(x{i}=x{i + 1}?1:0)+" for i in range(1, stations)] + [f"(x{stations}=x1?1:0);"])
    probs_const = ["\nconst double p0 = P_START;"] + \
                  [f"const double p{i} = p{i - 1}+P_STEP;" for i in range(1, dices + 1)]
    processes = [
        f"module process{i} = process1 [ x1=x{i}, y1=y{i}, x{stations}=x{i - 1}, m1=m{i}, step1=step{i} ] endmodule"
        for i in range(2, stations + 1)
    ]

    with open("workspace/simon/herman/custom/default.templ", "r") as default_templ:
        default_template = default_templ.read()
    if os.path.exists("workspace/simon/herman/custom/sketch.templ"):
        os.remove("workspace/simon/herman/custom/sketch.templ")
    with open("workspace/simon/herman/custom/sketch.templ", "w") as sketch_templ:
        sketch_templ.write(
            default_template + p_step + construct_holes(dices) + construct_controller(stations) +
            num_tokens + "\n" + "\n".join(probs_const) + "\n" +
            construct_process(dices, stations) + "\n" + "\n".join(processes)
        )


if __name__ == '__main__':
    generate_herman_sketch()



import json

class Fsc:
    '''
    An FSC having
    - a fixed number of nodes
    - a joint transition function of the type NxZ -> Distr(ActxN), where f(n,z) is a dictionary
        (action,memory) -> probability
    '''

    def __init__(self, num_nodes, num_observations):
        self.num_nodes = num_nodes
        self.num_observations = num_observations        
        self.transitions = [ [None]*num_observations for _ in range(num_nodes) ]
        self.observation_labels = None
        self.action_labels = None

    def check_transitions(self, observation_to_actions):
        assert len(self.transitions) == self.num_nodes, "FSC transition_function function is not defined for all memory nodes"
        for node in range(self.num_nodes):
            assert len(self.transitions[node]) == self.num_observations, \
                "in memory node {}, FSC transition function is not defined for all observations".format(node)
            for obs in range(self.num_observations):
                for action,update in self.transitions[node][obs].keys():
                    assert action in observation_to_actions[obs], "in observation {} FSC chooses invalid action {}".format(obs,action)
                    assert 0 <= update and update < self.num_nodes, "invalid FSC memory update {}".format(update)

    def check(self, observation_to_actions):
        ''' Check whether fields of FSC have been initialized appropriately. '''
        assert self.num_nodes > 0, "FSC must have at least 1 node"
        self.check_transitions(observation_to_actions)


class FscFactored:
    '''
    Class for encoding an FSC having
    - a fixed number of nodes
    - action selection is either:
        + deterministic: gamma: NxZ -> Act, or
        + randomized: gamma: NxZ -> Distr(Act), where gamma(n,z) is a dictionary of pairs (action,probability)
    - deterministic posterior-unaware memory update delta: NxZ -> N
    '''

    def __init__(self, num_nodes, num_observations, is_deterministic=False):
        self.num_nodes = num_nodes
        self.num_observations = num_observations
        self.is_deterministic = is_deterministic
        
        self.action_function = [ [None]*num_observations for _ in range(num_nodes) ]
        self.update_function = [ [None]*num_observations for _ in range(num_nodes) ]

        self.observation_labels = None
        self.action_labels = None

    def __str__(self):
        return json.dumps(self.to_json(), indent=4)

    def action_function_signature(self):
        if self.is_deterministic:
            return " NxZ -> Act"
        else:
            return " NxZ -> Distr(Act)"

    def to_json(self):
        json = {}
        json["num_nodes"] = self.num_nodes
        json["num_observations"] = self.num_observations
        json["__comment_action_function"] = "action function has signature {}".format(self.action_function_signature())
        json["__comment_update_function"] = "update function has signature NxZ -> N"

        json["action_function"] = self.action_function
        json["update_function"] = self.update_function

        if self.action_labels is not None:
            json["action_labels"] = self.action_labels
        if self.observation_labels is not None:
            json["observation_labels"] = self.observation_labels

        return json

    @classmethod
    def from_json(cls, json):
        num_nodes = json["num_nodes"]
        num_observations = json["num_observations"]
        fsc = FSC(num_nodes,num_observations)
        fsc.action_function = json["action_function"]
        fsc.update_function = json["update_function"]
        return fsc

    def reorder_nodes(self, node_old_to_new):
        action_function = [None for node in range(self.num_nodes)]
        update_function = [None for node in range(self.num_nodes)]
        for node_old,node_new in enumerate(node_old_to_new):
            action_function[node_new] = self.action_function[node_old]
            update_function[node_new] = [node_old_to_new[node] for node in self.update_function[node_old]]
        self.action_function = action_function
        self.update_function = update_function

    def reorder_actions(self, action_labels):
        for node in range(self.num_nodes):
            for obs in range(self.num_observations):
                if self.is_deterministic:
                    action = self.action_function[node][obs]
                    self.action_function[node][obs] = action_labels.index(self.action_labels[action])
                else:
                    action_function = {}
                    for action,prob in self.action_function[node][obs].items():
                        action_function[action_labels.index(self.action_labels[action])] = prob
                    self.action_function[node][obs] = action_function
        self.action_labels = action_labels.copy()

    def make_stochastic(self):
        if not self.is_deterministic:
            return
        for node in range(self.num_nodes):
            for obs in range(self.num_observations):
                self.action_function[node][obs] = {self.action_function[node][obs] : 1.0}
                self.update_function[node][obs] = {self.update_function[node][obs] : 1.0}
        self.is_deterministic = False

    def check_action_function(self, observation_to_actions):
        assert len(self.action_function) == self.num_nodes, "FSC action function is not defined for all memory nodes"
        for node in range(self.num_nodes):
            assert len(self.action_function[node]) == self.num_observations, \
                "in memory node {}, FSC action function is not defined for all observations".format(node)
            for obs in range(self.num_observations):
                if observation_to_actions[obs] == []:
                    assert self.action_function[node][obs] is None
                    continue
                if self.is_deterministic:
                    action_support = [self.action_function[node][obs]]
                else:
                    action_support = self.action_function[node][obs].keys()
                for action in action_support:
                    assert action in observation_to_actions[obs], "in observation {} FSC chooses invalid action {}".format(obs,action)

    def check_update_function(self):
        assert len(self.update_function) == self.num_nodes, "FSC update function is not defined for all memory nodes"
        for node in range(self.num_nodes):
            assert len(self.update_function[node]) == self.num_observations, \
                "in memory node {}, FSC update function is not defined for all observations".format(node)
            for obs in range(self.num_observations):
                update = self.update_function[node][obs]
                continue # skipping check due to inconsistency in our FSC definition
                assert 0 <= update and update < self.num_nodes, "invalid FSC memory update {}".format(update)

    def check(self, observation_to_actions):
        ''' Check whether fields of FSC have been initialized appropriately. '''
        assert self.num_nodes > 0, "FSC must have at least 1 node"
        self.check_action_function(observation_to_actions)
        self.check_update_function()

    def fill_trivial_actions(self, observation_to_actions):
        ''' For each observation with 1 available action, set gamma(n,z) to that action. '''
        for obs,actions in enumerate(observation_to_actions):
            if len(actions) != 1:
                continue
            action = actions[0]
            if not self.is_deterministic:
                action = {action:1}
            for node in range(self.num_nodes):
                self.action_function[node][obs] = action

    def fill_trivial_updates(self, observation_to_actions):
        ''' For each observation with 1 available action, set delta(n,z) to n. '''
        for obs,actions in enumerate(observation_to_actions):
            if len(actions)>1:
                continue
            for node in range(self.num_nodes):
                self.update_function[node][obs] = node

    def fill_zero_updates(self):
        for node in range(self.num_nodes):
            self.update_function[node] = [0] * self.num_observations

    def fill_implicit_actions_and_updates(self):
        '''
        For an FSC with an irregular memory model, make action and updates explicit.
        '''
        for node in range(self.num_nodes):
            for obs in range(self.num_observations):
                if self.action_function[node][obs] is None:
                    self.action_function[node][obs] = self.action_function[0][obs]
                    self.update_function[node][obs] = self.update_function[0][obs]

    def copy(self):
        fsc = FSC(self.num_nodes, self.num_observations, self.is_deterministic)
        fsc.action_function = [ [action for action in node] for node in self.action_function ]
        fsc.update_function = [ [update for update in node] for node in self.update_function ]
        fsc.observation_labels = self.observation_labels
        fsc.action_labels = self.action_labels
        return fsc



import paynt.quotient.quotient

import stormpy
import payntbind
import json
import graphviz

import logging
logger = logging.getLogger(__name__)

class Variable:

    def __init__(self, variable, name, domain):
        self.name = name
        # conversion of boolean variables to integers
        domain_new = []
        for value in domain:
            if value is True:
                value = 1
            elif value is False:
                value = 0
            domain_new.append(value)
        domain = domain_new
        domain = sorted(domain)
        self.domain = domain

    @property
    def domain_min(self):
        return self.domain[0]

    @property
    def domain_max(self):
        return self.domain[-1]

    @property
    def hole_domain(self):
        '''
        Hole domain does not include the maximum value.
        '''
        return self.domain[:-1]

    def __str__(self):
        # domain = "bool" if self.has_boolean_type else f"[{self.domain_min}..{self.domain_max}]"
        domain = f"[{self.domain_min}..{self.domain_max}]"
        return f"{self.name}:{domain}"




class DecisionTreeNode:

    def __init__(self, parent):
        self.parent = parent
        self.child_true = None
        self.child_false = None
        self.identifier = None
        self.holes = None

        self.action = None
        self.variable = None
        self.variable_bound = None

    @property
    def is_terminal(self):
        return self.child_false is None and self.child_true is None

    @property
    def child_nodes(self):
        return [] if self.is_terminal else [self.child_true,self.child_false]

    @property
    def is_true_child(self):
        return self is self.parent.child_true

    def add_children(self):
        assert self.is_terminal
        self.child_true = DecisionTreeNode(self)
        self.child_false = DecisionTreeNode(self)

    def get_depth(self):
        if self.is_terminal:
            return 0
        return 1 + max([child.get_depth() for child in self.child_nodes])

    def assign_identifiers(self, identifier=0):
        self.identifier = identifier
        if self.is_terminal:
            return self.identifier
        identifier = self.child_true.assign_identifiers(identifier+1)
        identifier = self.child_false.assign_identifiers(identifier+1)
        return identifier

    def associate_holes(self, node_hole_info):
        self.holes = [hole for hole,_,_ in node_hole_info[self.identifier]]
        if self.is_terminal:
            return
        self.child_true.associate_holes(node_hole_info)
        self.child_false.associate_holes(node_hole_info)

    def associate_assignment(self, assignment):
        hole_assignment = [assignment.hole_options(hole)[0] for hole in self.holes]
        if self.is_terminal:
            self.action = hole_assignment[0]
            return

        self.variable = hole_assignment[0]
        self.variable_bound = hole_assignment[self.variable+1]

        self.child_true.associate_assignment(assignment)
        self.child_false.associate_assignment(assignment)

    def apply_hint(self, subfamily, tree_hint):
        if self.is_terminal or tree_hint.is_terminal:
            return

        variable_hint = tree_hint.variable
        subfamily.hole_set_options(self.holes[0],[variable_hint])
        subfamily.hole_set_options(self.holes[variable_hint+1],[tree_hint.variable_bound])
        self.child_true.apply_hint(subfamily,tree_hint.child_true)
        self.child_false.apply_hint(subfamily,tree_hint.child_false)

    def simplify(self, variables, state_valuations):
        if self.is_terminal:
            return

        bound = variables[self.variable].domain[self.variable_bound]
        state_valuations_true =  [valuation for valuation in state_valuations if valuation[self.variable] <= bound]
        state_valuations_false = [valuation for valuation in state_valuations if valuation[self.variable]  > bound]
        child_skip = None
        if len(state_valuations_true) == 0:
            child_skip = self.child_false
        elif len(state_valuations_false) == 0:
            child_skip = self.child_true
        if child_skip is not None:
            self.variable = child_skip.variable
            self.variable_bound = child_skip.variable_bound
            self.action = child_skip.action
            self.child_true = child_skip.child_true
            self.child_false = child_skip.child_false
            self.simplify(variables,state_valuations)
            return

        self.child_true.simplify(variables, state_valuations_true)
        self.child_false.simplify(variables, state_valuations_false)
        if not self.is_terminal and self.child_true.is_terminal and self.child_false.is_terminal and self.child_true.action == self.child_false.action:
            self.variable = self.variable_bound = None
            self.action = self.child_true.action
            self.child_true = self.child_false = None

    def branch_expression(self, variables, true_branch=True):
        var = variables[self.variable]
        if true_branch:
            return f"{var.name}<={var.domain[self.variable_bound]}"
        else:
            return f"{var.name}>{var.domain[self.variable_bound]}"

    def path_expression(self, variables):
        if self.parent is None:
            return []
        return self.parent.path_expression(variables) + [self.parent.branch_expression(variables,true_branch=self.is_true_child)]

    def to_string(self, variables, action_labels, indent_level=0, indent_size=2):
        indent = " "*indent_level*indent_size
        if self.is_terminal:
            return indent + f"{action_labels[self.action]}" + "\n"
        s = ""
        s += indent + f"if {self.branch_expression(variables)}:" + "\n"
        s += self.child_true.to_string(variables,action_labels,indent_level+1)
        s += indent + f"else:" + "\n"
        s += self.child_false.to_string(variables,action_labels,indent_level+1)
        return s

    @property
    def graphviz_id(self):
        return str(self.identifier)

    def to_graphviz(self, graphviz_tree, variables, action_labels):
        if not self.is_terminal:
            for child in self.child_nodes:
                child.to_graphviz(graphviz_tree,variables,action_labels)

        if self.is_terminal:
            node_label = action_labels[self.action]
        else:
            var = variables[self.variable]
            node_label = f"{var.name}<={var.domain[self.variable_bound]}"

        graphviz_tree.node(self.graphviz_id, label=node_label, shape="box", style="rounded", margin="0.05,0.05")
        if not self.is_terminal:
            graphviz_tree.edge(self.graphviz_id,self.child_true.graphviz_id,label="T")
            graphviz_tree.edge(self.graphviz_id,self.child_false.graphviz_id,label="F")



class DecisionTree:

    def __init__(self, quotient, variables):
        self.quotient = quotient
        self.variables = variables
        self.reset()

    def reset(self):
        self.root = DecisionTreeNode(None)

    def set_depth(self, depth:int):
        self.reset()
        for level in range(depth):
            for node in self.collect_terminals():
                node.add_children()
        self.root.assign_identifiers()

    def get_depth(self):
        return self.root.get_depth()

    def collect_nodes(self, node_condition=None):
        if node_condition is None:
            node_condition = lambda node : True
        node_queue = [self.root]
        output_nodes = []
        while node_queue:
            node = node_queue.pop(0)
            if node_condition(node):
                output_nodes.append(node)
            node_queue += node.child_nodes
        return output_nodes

    def collect_terminals(self):
        return self.collect_nodes(lambda node : node.is_terminal)

    def collect_nonterminals(self):
        return self.collect_nodes(lambda node : not node.is_terminal)

    def to_list(self):
        num_nodes = len(self.collect_nodes())
        node_info = [ None for node in range(num_nodes) ]
        for node in self.collect_nodes():
            parent = num_nodes if node.parent is None else node.parent.identifier
            child_true = num_nodes if node.is_terminal else node.child_true.identifier
            child_false = num_nodes if node.is_terminal else node.child_false.identifier
            node_info[node.identifier] = (parent,child_true,child_false)
        return node_info

    def simplify(self, state_valuations):
        self.root.simplify(self.variables, state_valuations)

    def to_string(self):
        return self.root.to_string(self.variables,self.quotient.action_labels)

    def to_prism(self, indent_size=2):
        indent = " "*indent_size
        s = ""
        s += "module scheduler\n"
        for terminal in self.collect_terminals():
            action = f"{self.quotient.action_labels[terminal.action]}"
            guard = " & ".join(terminal.path_expression(self.variables))
            if guard == "":
                guard = "true"
            s += f"{indent}[{action}] {guard} -> true;\n"
        # s += indent + f"if {var.name}<={var.domain[self.variable_bound]}:" + "\n"
        # s += self.child_true.to_string(variables,action_labels,indent_level+1)
        # s += indent + f"else:" + "\n"
        # s += self.child_false.to_string(variables,action_labels,indent_level+1)
        s += "endmodule\n"
        return s

    def to_graphviz(self):
        logging.getLogger("graphviz").setLevel(logging.WARNING)
        logging.getLogger("graphviz.sources").setLevel(logging.ERROR)
        graphviz_tree = graphviz.Digraph(comment="decision tree")
        self.root.to_graphviz(graphviz_tree,self.variables,self.quotient.action_labels)
        return graphviz_tree


class MdpQuotient(paynt.quotient.quotient.Quotient):

    # label for action executing a random action selection
    DONT_CARE_ACTION_LABEL = "__random__"
    # if true, an explicit action executing a random choice of an available action will be added to each state
    add_dont_care_action = False
    # if true, irrelevant states will not be considered for tree mapping
    filter_deterministic_states = True

    @classmethod
    def get_state_valuations(cls, model):
        ''' Identify variable names and extract state valuation in the same order. '''
        assert model.has_state_valuations(), "model has no state valuations"
        # get name
        sv = model.state_valuations
        variable_name = None
        state_valuations = []
        for state in range(model.nr_states):
            valuation = json.loads(str(sv.get_json(state)))
            if variable_name is None:
                variable_name = list(valuation.keys())
            valuation = [valuation[var_name] for var_name in variable_name]
            state_valuations.append(valuation)
        return variable_name,state_valuations

    def __init__(self, mdp, specification):
        super().__init__(specification=specification)

        # mask of relevant states: non-absorbing states with more than one action
        self.state_is_relevant = None
        # bitvector of relevant states
        self.state_is_relevant_bv = None

        # list of relevant variables: variables having at least two different options on relevant states
        self.variables = None
        # for every state, a valuation of relevant variables
        self.relevant_state_valuations = None
        # decision tree obtained after reset_tree
        self.decision_tree = None

        # deprecated
        # updated = payntbind.synthesis.restoreActionsInAbsorbingStates(mdp)
        # if updated is not None: mdp = updated

        # identify relevant states
        self.state_is_relevant = [True for state in range(mdp.nr_states)]
        state_is_absorbing = self.identify_absorbing_states(mdp)
        self.state_is_relevant = [relevant and not state_is_absorbing[state] for state,relevant in enumerate(self.state_is_relevant)]

        if MdpQuotient.filter_deterministic_states:
            state_has_actions = self.identify_states_with_actions(mdp)
            self.state_is_relevant = [relevant and state_has_actions[state] for state,relevant in enumerate(self.state_is_relevant)]
        self.state_is_relevant_bv = stormpy.BitVector(mdp.nr_states)
        [self.state_is_relevant_bv.set(state,value) for state,value in enumerate(self.state_is_relevant)]
        logger.debug(f"MDP has {self.state_is_relevant_bv.number_of_set_bits()}/{self.state_is_relevant_bv.size()} relevant states")

        action_labels,_ = payntbind.synthesis.extractActionLabels(mdp)
        if MdpQuotient.DONT_CARE_ACTION_LABEL not in action_labels and MdpQuotient.add_dont_care_action:
            logger.debug("adding explicit don't-care action to relevant states...")
            mdp = payntbind.synthesis.addDontCareAction(mdp,self.state_is_relevant_bv)

        self.quotient_mdp = mdp
        self.choice_destinations = payntbind.synthesis.computeChoiceDestinations(mdp)
        self.action_labels,self.choice_to_action = payntbind.synthesis.extractActionLabels(mdp)
        logger.info(f"MDP has {len(self.action_labels)} actions")
        # TODO filter irrelevant actions?

        # get variable domains on relevant states
        variable_name,state_valuations = self.get_state_valuations(mdp)
        num_variables = len(variable_name)
        variable_domain = [set() for variable in range(num_variables)]
        for state in self.state_is_relevant_bv:
            valuation = state_valuations[state]
            for variable in range(num_variables):
                variable_domain[variable].add(valuation[variable])
        variable_domain = [sorted(domain) for domain in variable_domain]

        # filter variables having only one option
        variable_mask = [len(domain) > 1 for domain in variable_domain]
        variable_name = [value for variable,value in enumerate(variable_name) if variable_mask[variable]]
        variable_domain = [value for variable,value in enumerate(variable_domain) if variable_mask[variable]]
        # we filter unused variables from state valuations: this means that multiple states can now have the same "valuation"
        state_valuations = [
            [value for variable,value in enumerate(valuations) if variable_mask[variable]]
            for valuations in state_valuations
        ]

        self.variables = [Variable(variable,name,variable_domain[variable]) for variable,name in enumerate(variable_name)]
        self.relevant_state_valuations = state_valuations
        logger.debug(f"found the following {len(self.variables)} variables: {[str(v) for v in self.variables]}")


    def scheduler_json_to_choices(self, scheduler_json, discard_unreachable_states=False):
        variable_name,state_valuations = self.get_state_valuations(self.quotient_mdp)
        nci = self.quotient_mdp.nondeterministic_choice_indices.copy()
        assert self.quotient_mdp.nr_states == len(scheduler_json)
        state_to_choice = self.empty_scheduler()
        for state_decision in scheduler_json:
            valuation = [state_decision["s"][name] for name in variable_name]
            for state,state_valuation in enumerate(state_valuations):
                if valuation == state_valuation:
                    break
            else:
                assert False, "state valuation not found"

            actions = state_decision["c"]
            assert len(actions) == 1
            action_labels = actions[0]["labels"]
            assert len(action_labels) <= 1
            if len(action_labels) == 0:
                state_to_choice[state] = nci[state]
                continue
            action = self.action_labels.index(action_labels[0])
            # find a choice that executes this action
            for choice in range(nci[state],nci[state+1]):
                if self.choice_to_action[choice] == action:
                    state_to_choice[state] = choice
                    break
            else:
                assert False, "action is not available in the state"
        # enable implicit actions
        for state,choice in enumerate(state_to_choice):
            if choice is None:
                logger.warning(f"WARNING: scheduler has no action for state {state}")
                state_to_choice[state] = nci[state]

        if discard_unreachable_states:
            state_to_choice = self.discard_unreachable_choices(state_to_choice)
        # keep only relevant states
        state_to_choice = [choice if self.state_is_relevant[state] else None for state,choice in enumerate(state_to_choice)]
        choices = self.state_to_choice_to_choices(state_to_choice)

        scheduler_json_relevant = []
        for state_decision in scheduler_json:
            valuation = [state_decision["s"][name] for name in variable_name]
            for state,state_valuation in enumerate(state_valuations):
                if valuation == state_valuation:
                    break
            if state_to_choice[state] is None:
                continue
            scheduler_json_relevant.append(state_decision)

        return choices,scheduler_json_relevant
    
    # gets all choices that represent random action, used to compute the value of uniformly random scheduler
    def get_random_choices(self):
        nci = self.quotient_mdp.nondeterministic_choice_indices.copy()
        state_to_choice = self.empty_scheduler()
        random_action = self.action_labels.index(MdpQuotient.DONT_CARE_ACTION_LABEL)
        for state in range(self.quotient_mdp.nr_states):
            # find a choice that executes this action
            for choice in range(nci[state],nci[state+1]):
                if self.choice_to_action[choice] == random_action:
                    state_to_choice[state] = choice
                    break
        for state,choice in enumerate(state_to_choice):
            if choice is None:
                state_to_choice[state] = nci[state]

        choices = self.state_to_choice_to_choices(state_to_choice)

        return choices


    def reset_tree(self, depth, enable_harmonization=True):
        '''
        Rebuild the decision tree template, the design space and the coloring.
        '''
        logger.debug(f"building tree of depth {depth}")

        num_actions = len(self.action_labels)
        dont_care_action = num_actions
        if MdpQuotient.DONT_CARE_ACTION_LABEL in self.action_labels:
            dont_care_action = self.action_labels.index(MdpQuotient.DONT_CARE_ACTION_LABEL)

        self.decision_tree = DecisionTree(self,self.variables)
        self.decision_tree.set_depth(depth)

        variables = self.decision_tree.variables
        variable_name = [v.name for v in variables]
        variable_domain = [v.domain for v in variables]
        tree_list = self.decision_tree.to_list()
        self.coloring = payntbind.synthesis.ColoringSmt(
            self.quotient_mdp.nondeterministic_choice_indices, self.choice_to_action,
            num_actions, dont_care_action,
            self.quotient_mdp.state_valuations, self.state_is_relevant_bv,
            variable_name, variable_domain, tree_list, enable_harmonization
        )
        # return
        self.coloring.enableStateExploration(self.quotient_mdp)

        # reconstruct the family
        hole_info = self.coloring.getFamilyInfo()
        self.family = paynt.family.family.Family()
        self.is_action_hole = [False for hole in hole_info]
        self.is_decision_hole = [False for hole in hole_info]
        self.is_variable_hole = [False for hole in hole_info]
        domain_max = max([len(domain) for domain in variable_domain])
        bound_domain = list(range(domain_max))
        node_hole_info = [[] for node in self.decision_tree.collect_nodes()]
        for hole,info in enumerate(hole_info):
            node,hole_name,hole_type = info
            node_hole_info[node].append( (hole,hole_name,hole_type) )
            if hole_type == "__action__":
                self.is_action_hole[hole] = True
                option_labels = self.action_labels
            elif hole_type == "__decision__":
                self.is_decision_hole[hole] = True
                option_labels = variable_name
            else:
                self.is_variable_hole[hole] = True
                variable = variable_name.index(hole_type)
                option_labels = variables[variable].hole_domain
            self.family.add_hole(hole_name, option_labels)
        self.decision_tree.root.associate_holes(node_hole_info)


    def build_unsat_result(self):
        spec_result = paynt.verification.property_result.MdpSpecificationResult()
        spec_result.constraints_result = paynt.verification.property_result.ConstraintsResult([])
        spec_result.optimality_result = paynt.verification.property_result.MdpOptimalityResult(None)
        spec_result.evaluate(None)
        spec_result.can_improve = False
        return spec_result

    def build(self, family):
        if family.parent_info is None:
            choices = self.coloring.selectCompatibleChoices(family.family)
        else:
            choices = self.coloring.selectCompatibleChoices(family.family, family.parent_info.selected_choices)
        assert choices.number_of_set_bits() > 0

        # proceed as before
        family.selected_choices = choices
        family.mdp = self.build_from_choice_mask(choices)
        family.mdp.family = family


    def are_choices_consistent(self, choices, family):
        ''' Separate method for profiling purposes. '''
        consistent,hole_selection = self.coloring.areChoicesConsistent(choices, family.family)
        for hole,options in enumerate(hole_selection):
            assert len(options) == len(set(options)), str(hole_selection)
            for option in options:
                assert option in family.hole_options(hole), \
                f"option {option} for hole {hole} ({mdp.family.hole_name(hole)}) is not in the family"
        return consistent,hole_selection


    def scheduler_is_consistent(self, mdp, prop, result):
        ''' Get hole options involved in the scheduler selection. '''
        scheduler = result.scheduler
        assert scheduler.memoryless and scheduler.deterministic
        state_to_choice = self.scheduler_to_state_to_choice(mdp, scheduler)
        choices = self.state_to_choice_to_choices(state_to_choice)
        if self.specification.is_single_property:
            mdp.family.scheduler_choices = choices
        consistent,hole_selection = self.are_choices_consistent(choices, mdp.family)
        return hole_selection, consistent


    def scheduler_scores(self, mdp, prop, result, selection):
        inconsistent_assignments = {hole:options for hole,options in enumerate(selection) if len(options) > 1 }
        assert len(inconsistent_assignments) > 0, f"obtained selection with no inconsistencies: {selection}"
        inconsistent_action_holes = [(hole,options) for hole,options in inconsistent_assignments.items() if self.is_action_hole[hole]]
        inconsistent_decision_holes = [(hole,options) for hole,options in inconsistent_assignments.items() if self.is_decision_hole[hole]]
        inconsistent_variable_holes = [(hole,options) for hole,options in inconsistent_assignments.items() if self.is_variable_hole[hole]]

        # choose one splitter
        splitter = None
        # try action or decision holes first
        if len(inconsistent_action_holes) > 0:
            splitter = inconsistent_action_holes[0][0]
        elif len(inconsistent_decision_holes) > 0:
            splitter = inconsistent_decision_holes[0][0]
        else:
            splitter = inconsistent_variable_holes[0][0]
        assert splitter is not None, "splitter not set"
        # force the score of the selected splitter
        return {splitter:10}


    def split(self, family):

        mdp = family.mdp
        assert not mdp.is_deterministic

        # split family wrt last undecided result
        result = family.analysis_result.undecided_result()
        hole_assignments = result.primary_selection
        scores = self.scheduler_scores(mdp, result.prop, result.primary.result, result.primary_selection)

        splitters = self.holes_with_max_score(scores)
        splitter = splitters[0]
        if self.is_action_hole[splitter] or self.is_decision_hole[splitter]:
            assert len(hole_assignments[splitter]) > 1
            core_suboptions,other_suboptions = self.suboptions_enumerate(mdp, splitter, hole_assignments[splitter])
        else:
            subfamily_options = family.hole_options(splitter)

            # split in half
            index_split = len(subfamily_options)//2

            # split by inconsistent options
            option_1 = hole_assignments[splitter][0]; index_1 = subfamily_options.index(option_1)
            option_2 = hole_assignments[splitter][1]; index_2 = subfamily_options.index(option_2)
            index_split = index_2

            core_suboptions = [subfamily_options[:index_split], subfamily_options[index_split:]]

            for options in core_suboptions: assert len(options) > 0
            other_suboptions = []

        if len(other_suboptions) == 0:
            suboptions = core_suboptions
        else:
            suboptions = [other_suboptions] + core_suboptions  # DFS solves core first

        # construct corresponding subfamilies
        parent_info = family.collect_parent_info(self.specification)
        parent_info.analysis_result = family.analysis_result
        parent_info.scheduler_choices = family.scheduler_choices
        # parent_info.unsat_core_hint = self.coloring.unsat_core.copy()
        subfamilies = family.split(splitter,suboptions)
        assert family.size == sum([family.size for family in subfamilies])
        for subfamily in subfamilies:
            subfamily.add_parent_info(parent_info)
        return subfamilies



__version__ = "unknown"

try:
    from .._version import __version__
except ImportError:
    # We're running in a tree that doesn't have a _version.py, so we don't know what our version is.
    pass

def version():
    return __version__



import stormpy

from .statistic import Statistic
from .synthesizer_ar import SynthesizerAR
from .synthesizer_ar_storm import SynthesizerARStorm
from .synthesizer_hybrid import SynthesizerHybrid
from .synthesizer_multicore_ar import SynthesizerMultiCoreAR

import paynt.quotient.quotient
import paynt.quotient.decpomdp
import paynt.verification.property

import logging 
logger = logging.getLogger(__name__)



class SynthesizerDecPomdp:

    def __init__(self, quotient):
        self.quotient = quotient
        # TODO add support for more engines
        self.synthesizer = paynt.synthesizer.synthesizer_ar.SynthesizerAR
        self.total_iters = 0


    def synthesize(self, family, print_stats=True):
        synthesizer = self.synthesizer(self.quotient)   
        family.constraint_indices = self.quotient.family.constraint_indices
        assignment = synthesizer.synthesize(family, keep_optimum=True, print_stats=print_stats)
        self.total_iters += synthesizer.stat.iterations_mdp
        return assignment

    def strategy_iterative(self, unfold_imperfect_only):
        '''
        @param unfold_imperfect_only if True, only imperfect observations will be unfolded
        '''
            
        mem_size = paynt.quotient.decpomdp.DecPomdpQuotient.initial_memory_size
        opt = self.quotient.specification.optimality.optimum
        while True:
            
            logger.info("Synthesizing optimal k={} controller ...".format(mem_size))
            if mem_size > self.quotient.current_memory_size:
                if unfold_imperfect_only:
                    self.quotient.set_imperfect_memory_size(mem_size)
                else:
                    self.quotient.set_global_memory_size(mem_size)
            
            self.synthesize(self.quotient.family)

            opt_old = opt
            opt = self.quotient.specification.optimality.optimum

            mem_size += 1

            #break


    def run(self, optimum_threshold=None):
        self.strategy_iterative(unfold_imperfect_only=True)



import time
import psutil
import os

class Timer:

    def __init__(self, time_limit_seconds=None):
        self.running = False
        self.last_timestamp = None
        self.time = 0 # total time measured
        self.time_limit_seconds = time_limit_seconds

    @staticmethod
    def timestamp():
        return time.perf_counter()

    def reset(self):
        self.__init__()

    def start(self):
        if self.running:
            return
        self.last_timestamp = self.timestamp()
        self.running = True

    def stop(self):
        if not self.running:
            return
        self.time += self.timestamp() - self.last_timestamp
        self.last_timestamp = None
        self.running = False

    def read(self):
        if not self.running:
            return self.time
        else:
            return self.time + (self.timestamp() - self.last_timestamp)

    def time_limit_reached(self):
        return self.time_limit_seconds is not None and self.read() > self.time_limit_seconds


class GlobalTimer:

    global_timer = None

    @classmethod
    def start(cls, time_limit_seconds=None):
        cls.global_timer = Timer(time_limit_seconds)
        cls.global_timer.start()

    @classmethod
    def read(cls):
        if cls.global_timer is None:
            logger.warning("attempting to read uninitialized GlobalTimer")
            return -1
        return cls.global_timer.read()

    @classmethod
    def time_limit_reached(cls):
        return cls.global_timer is not None and cls.global_timer.time_limit_reached()


class GlobalMemoryLimit:

    memory_limit_mb = None

    @classmethod
    def limit_reached(cls):
        process = psutil.Process(os.getpid())
        allocated_mb = process.memory_info().rss / (1024 * 1024)
        return cls.memory_limit_mb is not None and allocated_mb > cls.memory_limit_mb



import payntbind.synthesis

import paynt.family.smt

import math
import random
import itertools

import logging
logger = logging.getLogger(__name__)


class ParentInfo():
    '''
    Container for stuff to be remembered when splitting an undecided family into subfamilies. Generally used to
    speed-up work with the subfamilies.
    :note it is better to store these things in a separate container instead
        of having a reference to the parent family (that will never be considered again) for memory efficiency.
    '''
    def __init__(self):
        pass
        self.selected_choices = None
        self.constraint_indices = None
        self.refinement_depth = None


class Family:

    def __init__(self, other=None):
        if other is None:
            self.family = payntbind.synthesis.Family()
            self.hole_to_name = []
            self.hole_to_option_labels = []
        else:
            self.family = payntbind.synthesis.Family(other.family)
            self.hole_to_name = other.hole_to_name
            self.hole_to_option_labels = other.hole_to_option_labels

        self.parent_info = None
        self.refinement_depth = 0
        self.constraint_indices = None

        self.selected_choices = None
        self.mdp = None
        self.analysis_result = None
        self.encoding = None

    def add_parent_info(self, parent_info):
        self.parent_info = parent_info
        self.refinement_depth = parent_info.refinement_depth + 1
        self.constraint_indices = parent_info.constraint_indices

    @property
    def num_holes(self):
        return self.family.numHoles()

    def add_hole(self, name, option_labels):
        self.hole_to_name.append(name)
        self.hole_to_option_labels.append(option_labels)
        self.family.addHole(len(option_labels))

    def hole_name(self, hole):
        return self.hole_to_name[hole]

    def hole_options(self, hole):
        return self.family.holeOptions(hole)

    def hole_num_options(self, hole):
        return self.family.holeNumOptions(hole)

    def hole_num_options_total(self, hole):
        return self.family.holeNumOptionsTotal(hole)

    def hole_set_options(self, hole, options):
        self.family.holeSetOptions(hole,options)

    @property
    def size(self):
        return math.prod([self.family.holeNumOptions(hole) for hole in range(self.num_holes)])

    INT_PRINT_MAX_ORDER = 5

    @property
    def size_or_order(self):
        order = int(math.fsum([math.log10(self.family.holeNumOptions(hole)) for hole in range(self.num_holes)]))
        if order <= Family.INT_PRINT_MAX_ORDER:
            return self.size
        return f"1e{order}"

    def hole_options_to_string(self, hole, options):
        name = self.hole_name(hole)
        labels = [str(self.hole_to_option_labels[hole][option]) for option in options]
        if len(labels) == 1:
            return f"{name}={labels[0]}"
        else:
            return name + ": {" + ",".join(labels) + "}"

    def __str__(self):
        hole_strings = []
        for hole in range(self.num_holes):
            options = self.hole_options(hole)
            hole_str = self.hole_options_to_string(hole,options)
            hole_strings.append(hole_str)
        return ", ".join(hole_strings)

    def copy(self):
        return Family(self)

    def assume_hole_options_copy(self, hole, options):
        '''
        Create a copy and assume suboptions for a given hole.
        @note this does not check whether @options are actually suboptions of this hole.
        '''
        subfamily = self.copy()
        subfamily.hole_set_options(hole,options)
        return subfamily

    def assume_options_copy(self, hole_options):
        '''
        Create a copy and assume suboptions for each hole.
        @note this does not check whether suboptions are actually suboptions of any given hole.
        '''
        subfamily = self.copy()
        for hole,options in enumerate(hole_options):
            subfamily.hole_set_options(hole,options)
        return subfamily

    def split(self, splitter, suboptions):
        return [self.assume_hole_options_copy(splitter,options) for options in suboptions]

    def pick_any(self):
        hole_options = [[self.hole_options(hole)[0]] for hole in range(self.num_holes)]
        return self.assume_options_copy(hole_options)

    def pick_random(self):
        hole_options = [[random.choice(self.hole_options(hole))] for hole in range(self.num_holes)]
        return self.assume_options_copy(hole_options)

    def all_combinations(self):
        '''
        :returns iteratable Cartesian product of hole options
        '''
        all_options = []
        for hole in range(self.num_holes):
            options = self.hole_options(hole)
            all_options.append(options)
        return itertools.product(*all_options)

    def construct_assignment(self, combination):
        ''' Convert hole option combination to a hole assignment. '''
        combination = list(combination)
        suboptions = [[option] for option in combination]
        assignment = self.assume_options_copy(suboptions)
        return assignment

    def collect_parent_info(self, specification):
        pi = ParentInfo()
        pi.selected_choices = self.selected_choices
        pi.refinement_depth = self.refinement_depth
        cr = self.analysis_result.constraints_result
        pi.constraint_indices = cr.undecided_constraints if cr is not None else []
        return pi

    def encode(self, smt_solver):
        if self.encoding is None:
            self.encoding = paynt.family.smt.FamilyEncoding(smt_solver, self)



import stormpy
import payntbind

import paynt.utils.timer

import logging
logger = logging.getLogger(__name__)

class AllInOne:

    def __init__(self, all_in_one_program, specification, approach, memory_limit_mb, family):

        self.properties = specification.all_properties()
        self.threshold = self.properties[0].threshold
        logger.info("changing constraints into optimality") # TODO check if the properties are optimality or not
        self.transformed_properties = [x.transform_to_optimality_formula(all_in_one_program) for x in self.properties]
        self.approach = approach
        self.family = family

        build_timer = paynt.utils.timer.Timer()
        logger.info(f"building {self.approach} all in one MDP")

        build_timer.start()
        if self.approach == 'bdd':
            stormpy.set_settings(["--sylvan:maxmem", str(memory_limit_mb)]) # set memory usage by the symbolic approach
            self.model = stormpy.build_symbolic_model(all_in_one_program)
            self.filter = stormpy.create_filter_initial_states_symbolic(self.model)
        elif self.approach == 'sparse':
            build_options = stormpy.BuilderOptions([p.raw_formula for p in self.transformed_properties])
            build_options.set_build_state_valuations()
            self.model = stormpy.build_sparse_model_with_options(all_in_one_program, build_options)
            self.filter = stormpy.create_filter_initial_states_sparse(self.model)
        else:
            logger.error("Unknown all in one approach!")
            exit(1)
        build_timer.stop()

        logger.info(f"build finished in {round(build_timer.read(),1)}s")
        logger.info(f"constructed all in one MDP having {self.model.nr_states} states and {self.model.nr_choices} actions")


    def synthesize(self, optimum_threshold=None):

        all_in_one_timer = paynt.utils.timer.Timer()
        logger.info(f"starting all in one analysis")

        all_in_one_timer.start()
        if self.approach == 'bdd':
            result = stormpy.check_model_dd(self.model, self.transformed_properties[0], only_initial_states=True)
            result.filter(self.filter)
            all_in_one_timer.stop()

            members_sat = 0
            result_values = result.get_values()
            for valuation, mc_result in result_values:
                # 'valuation' can be used to obtain hole assignments for a give result 
                if self.properties[0].satisfies_threshold(mc_result):
                    members_sat += 1
                    
        elif self.approach == 'sparse':
            result = stormpy.check_model_sparse(self.model, self.transformed_properties[0], extract_scheduler=True, only_initial_states=True)
            result.filter(self.filter)
            all_in_one_timer.stop()
        
        members_total = self.family.size
        members_sat_percentage = int(round(members_sat/members_total*100,0))
        print(f"satisfied {members_sat}/{members_total} members ({members_sat_percentage}%)")

        time_elapsed = round(all_in_one_timer.read(),1)
        logger.info(f"all in one analysis finished in {time_elapsed}s")





from stormpy import pomdp
import paynt.synthesizer.synthesizer_ar
from ..quotient.storm_pomdp_control import StormPOMDPControl
from os import makedirs


from time import sleep

import logging
logger = logging.getLogger(__name__)

# Abstraction Refinement + Storm splitting
class SynthesizerARStorm(paynt.synthesizer.synthesizer_ar.SynthesizerAR):

    # family exploration order: True = DFS, False = BFS
    exploration_order_dfs = True

    # buffer containing subfamilies to be checked after the main restricted family
    subfamilies_buffer = None

    main_family = None

    # if True, Storm over-approximation will be run to help with family pruning
    storm_pruning = False

    storm_control = None
    s_queue = None

    saynt_timer = None

    @property
    def method_name(self):
        return "AR"

    # performs splitting in family according to Storm result
    # main families contain only those actions that were considered by best found Storm FSC
    def storm_split(self, families):
        subfamilies = []
        main_families = []

        # split each family in the current buffer to main family and corresponding subfamilies
        for family in families:
            if self.storm_control.use_cutoffs:
                result_dict = self.storm_control.result_dict
            else:
                result_dict = self.storm_control.result_dict_no_cutoffs
            main_p = self.storm_control.get_main_restricted_family(family, result_dict)
            if main_p is None:
                subfamilies.append(family)
                continue

            main_families.append(main_p)
            subfamily_restrictions = self.storm_control.get_subfamilies_restrictions(family, result_dict)
            subfamilies_p = self.storm_control.get_subfamilies(subfamily_restrictions, family)
            subfamilies.extend(subfamilies_p)

        logger.info(f"State after Storm splitting: Main families - {len(main_families)}, Subfamilies - {len(subfamilies)}")

        # if there are no main families we don't have to prioritize search
        if len(main_families) == 0:
            main_families = subfamilies
            subfamilies = []

        return main_families, subfamilies




    def verify_family(self, family):
        self.quotient.build(family)
        self.check_specification(family)

        if family.analysis_result.improving_value is not None:
            if self.saynt_timer is not None:
                print(f'-----------PAYNT----------- \
                    \nValue = {family.analysis_result.improving_value} | Time elapsed = {round(self.saynt_timer.read(),1)}s | FSC size = {self.quotient.policy_size(family.analysis_result.improving_assignment)}\n', flush=True)
            else:
                self.stat.new_fsc_found(family.analysis_result.improving_value, family.analysis_result.improving_assignment, self.quotient.policy_size(family.analysis_result.improving_assignment))
            self.quotient.specification.optimality.update_optimum(family.analysis_result.improving_value)

        # storm pruning runs Storm POMDP over-approximation analysis and on the sub-POMDP given by a family
        # this serves as a better abstraction for pruning, however is much more computationally intensive
        if self.quotient.specification.optimality.optimum and family.analysis_result.can_improve and self.storm_pruning:

            family_pomdp = self.quotient.get_family_pomdp(family.mdp)

            storm_res = StormPOMDPControl.storm_pomdp_analysis(family_pomdp, self.quotient.specification.stormpy_formulae())

            # compare computed bounds to the current optimum to see if the family can be pruned
            if self.quotient.specification.optimality.minimizing:
                if self.quotient.specification.optimality.optimum <= storm_res.lower_bound:
                    family.analysis_result.can_improve = False
                    logger.info(f"Used Storm result to prune a family with Storm value: {storm_res.lower_bound} compared to current optimum {self.quotient.specification.optimality.optimum}. Quotient MDP value: {res.optimality_result.primary.value}")
            else:
                if self.quotient.specification.optimality.optimum >= storm_res.upper_bound:
                    family.analysis_result.can_improve = False
                    logger.info(f"Used Storm result to prune a family with Storm value: {storm_res.upper_bound} compared to current optimum {self.quotient.specification.optimality.optimum}. Quotient MDP value: {res.optimality_result.primary.value}")



    def synthesize_one(self, family):

        self.best_assignment = None

        if self.main_family is not None:
            family = self.main_family

        families = [family]

        while families:

            # check whether PAYNT should be paused
            if self.s_queue is not None:
                # if the queue is non empty, pause for PAYNT was requested
                if not self.s_queue.empty():
                    if self.best_assignment is not None:
                        self.storm_control.latest_paynt_result = self.best_assignment
                        self.storm_control.paynt_export = self.quotient.extract_policy(self.best_assignment)
                        self.storm_control.paynt_bounds = self.quotient.specification.optimality.optimum
                        self.storm_control.paynt_fsc_size = self.quotient.policy_size(self.storm_control.latest_paynt_result)
                        self.storm_control.latest_paynt_result_fsc = self.quotient.assignment_to_fsc(self.storm_control.latest_paynt_result)
                        self.storm_control.update_data()
                    logger.info("Pausing synthesis")
                    self.s_queue.get()
                    self.stat.synthesis_timer.stop()
                    # check for the signal that PAYNT can be resumed or terminated
                    while self.s_queue.empty():
                        sleep(1)
                    status = self.s_queue.get()
                    if status == "resume":
                        logger.info("Resuming synthesis")
                        if self.storm_control.is_storm_better:
                            # if the result found by Storm is better and needs more memory end the current synthesis and add memory
                            if self.storm_control.is_memory_needed():
                                logger.info("Additional memory needed")
                                return self.best_assignment
                            else:
                                logger.info("Applying family split according to Storm results")
                                families, self.subfamilies_buffer = self.storm_split(families)
                        # if Storm's result is not better continue with the synthesis normally
                        else:
                            logger.info("PAYNT's value is better. Prioritizing synthesis results")
                        self.stat.synthesis_timer.start()

                    elif status == "terminate":
                        logger.info("Terminating controller synthesis")
                        return self.best_assignment

            if SynthesizerARStorm.exploration_order_dfs:
                family = families.pop(-1)
            else:
                family = families.pop(0)

            # simulate sequential
            family.parent_info = None

            self.verify_family(family)
            if family.analysis_result.improving_assignment is not None:
                self.best_assignment = family.analysis_result.improving_assignment
            # family can be pruned
            if family.analysis_result.can_improve == False:
                self.explore(family)
                # if there are no more families in the main buffer coninue the exploration in the subfamilies
                if not families and self.subfamilies_buffer:
                    logger.info("Main family synthesis done")
                    logger.info(f"Subfamilies buffer contains: {len(self.subfamilies_buffer)} families")
                    families = self.subfamilies_buffer
                    self.subfamilies_buffer = []
                continue

            # undecided
            subfamilies = self.quotient.split(family)
            families = families + subfamilies

        return self.best_assignment




import stormpy
import payntbind

import paynt.verification.property
import paynt.models.model_builder

import itertools
from collections import defaultdict

import logging
logger = logging.getLogger(__name__)


class CombinationColoring:
    '''
    Dictionary of colors associated with different hole combinations.
    Note: color 0 is reserved for general hole-free objects.
    '''
    def __init__(self):
        self.coloring = {}
        self.reverse_coloring = [None]

    @property
    def num_colors(self):
        return len(self.coloring)

    def get_or_make_color(self, hole_assignment):
        new_color = self.num_colors + 1
        color = self.coloring.get(hole_assignment, new_color)
        if color == new_color:
            self.coloring[hole_assignment] = color
            self.reverse_coloring.append(hole_assignment)
        return color


class JaniUnfolder:
    ''' Unfolder of hole combinations into JANI program. '''

    def __init__(self, prism, hole_expressions, specification, family):

        logger.debug("constructing JANI program...")
        
        # pack properties and translate Prism to Jani
        properties_old = specification.all_properties()
        stormpy_properties = [p.property for p in properties_old]
        jani,properties = prism.to_jani(stormpy_properties)

        # upon translation, some properties may change their atoms, so we need to re-wrap all properties
        properties_unpacked = []
        for index,prop_old in enumerate(properties_old):
            prop_new = properties[index]
            if type(prop_old) == paynt.verification.property.Property:
                p = paynt.verification.property.Property(prop_new)
            else:
                epsilon = prop_old.epsilon
                p = paynt.verification.property.OptimalityProperty(prop_new,epsilon)
            properties_unpacked.append(p)
        self.specification = paynt.verification.property.Specification(properties_unpacked)
        self.jani_unfolded,edge_to_hole_options = JaniUnfolder.unfold_jani(jani, family, hole_expressions)

        logger.debug("constructing the quotient...")
        quotient_mdp = paynt.models.model_builder.ModelBuilder.from_jani(self.jani_unfolded, self.specification)

        # associate each action of a quotient MDP with hole options
        # reconstruct choice labels from choice origins
        logger.debug("associating choices of the quotient with hole assignments...")
        choice_is_valid,choice_to_hole_options = payntbind.synthesis.janiMapChoicesToHoleAssignments(
            quotient_mdp,family.family,edge_to_hole_options
        )

        # handle conflicting colors
        num_choices_all = quotient_mdp.nr_choices
        num_choices_valid = choice_is_valid.number_of_set_bits()
        if num_choices_valid < num_choices_all:
            logger.debug("keeping {}/{} choices with non-conflicting hole assignments...".format(num_choices_valid,num_choices_all))
            keep_unreachable_states = False
            subsystem_builder_options = stormpy.SubsystemBuilderOptions()
            subsystem_builder_options.build_action_mapping = True
            all_states = stormpy.storage.BitVector(quotient_mdp.nr_states, True)
            submodel_construction = stormpy.construct_submodel(
                quotient_mdp, all_states, choice_is_valid, keep_unreachable_states, subsystem_builder_options
            )
            quotient_mdp = submodel_construction.model
            choice_map = list(submodel_construction.new_to_old_action_mapping)
            choice_to_hole_options = [choice_to_hole_options[choice_map[choice]] for choice in range(quotient_mdp.nr_choices)]

        self.quotient_mdp = quotient_mdp
        self.choice_to_hole_options = choice_to_hole_options
        return

    @staticmethod
    def unfold_jani(jani, family, hole_expressions):
        # ensure that jani.constants are in the same order as our holes
        open_constants = [c for c in jani.constants if not c.defined]
        hole_variables = [c.expression_variable for c in open_constants]
        assert len(open_constants) == family.num_holes
        for hole in range(family.num_holes):
            assert family.hole_name(hole) == open_constants[hole].name

        combination_coloring = CombinationColoring()
        jani_program = stormpy.JaniModel(jani)
        new_automata = dict()
        for aut_index,automaton in enumerate(jani_program.automata):
            if not JaniUnfolder.automaton_has_holes(automaton, hole_variables):
                continue
            new_aut = JaniUnfolder.construct_automaton(automaton, hole_variables, hole_expressions, combination_coloring)
            new_automata[aut_index] = new_aut
        for aut_index,aut in new_automata.items():
            jani_program.replace_automaton(aut_index, aut)
        for hole in range(family.num_holes):
            jani_program.remove_constant(family.hole_name(hole))

        jani_program.set_model_type(stormpy.JaniModelType.MDP)
        jani_program.finalize()
        jani_program.check_valid()

        # collect label and color of each edge
        edge_to_hole_options = {}
        for aut_index, automaton in enumerate(jani_program.automata):
            for edge_index, edge in enumerate(automaton.edges):
                if edge.color == 0:
                    continue
                global_index = jani_program.encode_automaton_and_edge_index(aut_index, edge_index)
                options = combination_coloring.reverse_coloring[edge.color]
                options = [(hole_index,option) for hole_index,option in enumerate(options) if option is not None]
                edge_to_hole_options[global_index] = options

        return jani_program,edge_to_hole_options

    @staticmethod
    def edge_holes(edge, hole_variables):
        variables = set()
        variables |= edge.guard.get_variables()
        for assignment in edge.template_edge.assignments:
            variables |= assignment.expression.get_variables()
        for dest in edge.destinations:
            variables |= dest.probability.get_variables()
            for assignment in dest.assignments:
                variables |= assignment.expression.get_variables()
        for dest in edge.template_edge.destinations:
            for assignment in dest.assignments:
                variables |= assignment.expression.get_variables()
        return [hole for hole,variable in enumerate(hole_variables) if variable in variables]

    @staticmethod
    def automaton_has_holes(automaton, hole_variables):
        for edge in automaton.edges:
            if len(JaniUnfolder.edge_holes(edge,hole_variables)) > 0:
                return True
        return False

    @staticmethod
    def construct_automaton(automaton, hole_variables, hole_expressions, combination_coloring):
        new_aut = stormpy.storage.JaniAutomaton(automaton.name, automaton.location_variable)
        [new_aut.add_location(loc) for loc in automaton.locations]
        [new_aut.add_initial_location(idx) for idx in automaton.initial_location_indices]
        [new_aut.variables.add_variable(var) for var in automaton.variables]
        for edge in automaton.edges:
            new_edges = JaniUnfolder.construct_edges(edge, hole_variables, hole_expressions, combination_coloring)
            for new_edge in new_edges:
                new_aut.add_edge(new_edge)
        return new_aut

    @staticmethod
    def construct_edges(edge, hole_variables, hole_expressions, combination_coloring):
        edge_holes = JaniUnfolder.edge_holes(edge,hole_variables)
        if len(edge_holes) == 0:
            return [JaniUnfolder.construct_edge(edge)]

        combinations = [
            (list(range(len(expressions))) if hole in edge_holes else [None])
            for hole,expressions in enumerate(hole_expressions)
        ]
        new_edges = []
        for combination in itertools.product(*combinations):
            substitution = {
                hole_variables[hole] : expressions[combination[hole]]
                for hole,expressions in enumerate(hole_expressions)
                if combination[hole] is not None
            }
            new_edge = JaniUnfolder.construct_edge(edge,substitution)
            new_edge.color = combination_coloring.get_or_make_color(combination)
            new_edges.append(new_edge)
        return new_edges

    @staticmethod
    def construct_edge(edge, substitution = None):
        guard = stormpy.Expression(edge.template_edge.guard)
        assignments = edge.template_edge.assignments.clone()
        if substitution is not None:
            guard = guard.substitute(substitution)
            assignments.substitute(substitution,substitute_transcendental_numbers=True)
        template_edge = stormpy.storage.JaniTemplateEdge(guard)
        payntbind.synthesis.janiTemplateEdgeAddAssignments(template_edge,assignments)
        for dst in edge.template_edge.destinations:
            assignments = dst.assignments.clone()
            if substitution is not None:
                assignments.substitute(substitution,substitute_transcendental_numbers=True)
            template_edge.add_destination(stormpy.storage.JaniTemplateEdgeDestination(assignments))

        destinations = [(dst.target_location_index,dst.probability) for dst in edge.destinations]
        if substitution is not None:
            destinations = [(target,prob.substitute(substitution)) for target,prob in destinations]
        return stormpy.storage.JaniEdge(
            edge.source_location_index, edge.action_index, edge.rate, template_edge, destinations
        )


    def write_jani(self, output_path):
        logger.debug(f"Writing unfolded program to {output_path}")
        with open(output_path, "w") as f:
            f.write(str(self.jani_unfolded))



import os
import stormpy
import stormpy.examples
import stormpy.examples.files

from paynt.examples import paynt_models_dir

stormpy_example_dir = stormpy.examples.files.testfile_dir


def get_stormpy_example_path(*paths):
    return os.path.join(stormpy_example_dir, *paths)

def get_builder_options():
    builder_options = stormpy.BuilderOptions()
    builder_options.set_build_state_valuations(True)
    builder_options.set_build_with_choice_origins(True)
    builder_options.set_build_all_labels(True)
    builder_options.set_build_choice_labels(True)
    builder_options.set_add_overlapping_guards_label(True)
    builder_options.set_build_observation_valuations(True)
    builder_options.set_build_all_reward_models(True)
    return builder_options

def get_sketch_paths(project_path, sketch_name="sketch.templ", props_name="sketch.props"):
    return os.path.join(paynt_models_dir, project_path, sketch_name), os.path.join(paynt_models_dir, project_path, props_name)

def read_first_line(filepath):
    with open(filepath, 'r') as file:
        first_line = file.readline().strip()
    return first_line


import paynt.quotient.mdp_family
from . import version

import paynt.utils.timer
import paynt.utils.version_check
import paynt.parser.sketch

import paynt.quotient.quotient
import paynt.quotient.pomdp
import paynt.quotient.decpomdp
import paynt.quotient.posmg
import paynt.quotient.storm_pomdp_control
import paynt.quotient.mdp

import paynt.synthesizer.synthesizer
import paynt.synthesizer.synthesizer_cegis
import paynt.synthesizer.policy_tree
import paynt.synthesizer.decision_tree

import click
import sys
import os
import cProfile, pstats

import logging
logger = logging.getLogger(__name__)


def setup_logger(log_path = None):
    ''' Setup routine for logging. '''

    root = logging.getLogger()
    root.setLevel(logging.DEBUG)
    # root.setLevel(logging.INFO)

    # formatter = logging.Formatter('%(asctime)s %(threadName)s - %(name)s - %(levelname)s - %(message)s')
    formatter = logging.Formatter('%(asctime)s - %(filename)s:%(lineno)d - %(message)s')

    handlers = []
    if log_path is not None:
        fh = logging.FileHandler(log_path)
        fh.setLevel(logging.DEBUG)
        fh.setFormatter(formatter)
        handlers.append(fh)
    sh = logging.StreamHandler(sys.stdout)
    handlers.append(sh)
    sh.setLevel(logging.DEBUG)
    sh.setFormatter(formatter)
    for h in handlers:
        root.addHandler(h)
    return handlers


@click.command()
@click.argument('project', type=click.Path(exists=True))
@click.option("--sketch", default="sketch.templ", show_default=True,
    help="name of the sketch file in the project")
@click.option("--props", default="sketch.props", show_default=True,
    help="name of the properties file in the project")
@click.option("--relative-error", type=click.FLOAT, default="0", show_default=True,
    help="relative error for optimal synthesis")
@click.option("--optimum-threshold", type=click.FLOAT,
    help="known optimum bound")
@click.option("--precision", type=click.FLOAT, default=1e-4,
    help="model checking precision")
@click.option("--exact", is_flag=True, default=False,
    help="use exact synthesis (very limited at the moment)")
@click.option("--timeout", type=int,
    help="timeout (s)")

@click.option("--export",
    type=click.Choice(['jani', 'drn', 'pomdp']),
    help="export the model to specified format and abort")

@click.option("--method",
    type=click.Choice(['onebyone', 'ar', 'cegis', 'hybrid', 'ar_multicore']),
    default="ar", show_default=True,
    help="synthesis method"
    )

@click.option("--disable-expected-visits", is_flag=True, default=False,
    help="do not compute expected visits for the splitting heuristic")

@click.option("--fsc-synthesis", is_flag=True, default=False,
    help="enable incremental synthesis of FSCs for a (Dec-)POMDP")
@click.option("--fsc-memory-size", default=1, show_default=True,
    help="implicit memory size for (Dec-)POMDP FSCs")
@click.option("--posterior-aware", is_flag=True, default=False,
    help="unfold MDP taking posterior observation of into account")

@click.option("--storm-pomdp", is_flag=True, default=False,
    help="enable running belief analysis in STorm to enhance FSC synthesis for POMDPs (AR only)")
@click.option(
    "--storm-options",
    default="cutoff",
    type=click.Choice(["cutoff", "clip2", "clip4", "small", "refine", "overapp", "2mil", "5mil", "10mil", "20mil", "30mil", "50mil"]),
    show_default=True,
    help="run Storm using pre-defined settings and use the result to enhance PAYNT. Can only be used together with --storm-pomdp flag")
@click.option("--iterative-storm", nargs=3, type=int, show_default=True, default=None,
    help="runs the iterative PAYNT/Storm integration. Arguments timeout, paynt_timeout, storm_timeout. Can only be used together with --storm-pomdp flag")
@click.option("--get-storm-result", default=None, type=int,
    help="runs PAYNT for given amount of seconds and returns Storm result using FSC at cutoff. If time is 0 returns pure Storm result. Can only be used together with --storm-pomdp flag")
@click.option("--prune-storm", is_flag=True, default=False,
    help="only explore the main family suggested by Storm in each iteration. Can only be used together with --storm-pomdp flag. Can only be used together with --storm-pomdp flag")
@click.option("--use-storm-cutoffs", is_flag=True, default=False,
    help="use storm randomized scheduler cutoffs are used during the prioritization of families. Can only be used together with --storm-pomdp flag. Can only be used together with --storm-pomdp flag")
@click.option(
    "--unfold-strategy-storm",
    default="storm",
    type=click.Choice(["storm", "paynt", "cutoff"]),
    show_default=True,
    help="specify memory unfold strategy. Can only be used together with --storm-pomdp flag")

@click.option("--export-synthesis", type=click.Path(), default=None,
    help="base filename to output synthesis result")

@click.option("--mdp-discard-unreachable-choices", is_flag=True, default=False,
    help="if set, unreachable choices will be discarded from the splitting scheduler")

@click.option("--tree-depth", default=0, type=int,
    help="decision tree synthesis: tree depth")
@click.option("--tree-enumeration", is_flag=True, default=False,
    help="decision tree synthesis: if set, all trees of size at most tree_depth will be enumerated")
@click.option("--tree-map-scheduler", type=click.Path(), default=None,
    help="decision tree synthesis: path to a scheduler to be mapped to a decision tree")
@click.option("--add-dont-care-action", is_flag=True, default=False,
    help="decision tree synthesis: # if set, an explicit action executing a random choice of an available action will be added to each state")

@click.option(
    "--constraint-bound", type=click.FLOAT, help="bound for creating constrained POMDP for Cassandra models",
)

@click.option(
    "--ce-generator", type=click.Choice(["dtmc", "mdp"]), default="dtmc", show_default=True,
    help="counterexample generator",
)
@click.option("--profiling", is_flag=True, default=False,
    help="run profiling")

def paynt_run(
    project, sketch, props, relative_error, optimum_threshold, precision, exact, timeout,
    export,
    method,
    disable_expected_visits,
    fsc_synthesis, fsc_memory_size, posterior_aware,
    storm_pomdp, iterative_storm, get_storm_result, storm_options, prune_storm,
    use_storm_cutoffs, unfold_strategy_storm,
    export_synthesis,
    mdp_discard_unreachable_choices,
    tree_depth, tree_enumeration, tree_map_scheduler, add_dont_care_action,
    constraint_bound,
    ce_generator,
    profiling
):

    profiler = None
    if profiling:
        profiler = cProfile.Profile()
        profiler.enable()
    paynt.utils.timer.GlobalTimer.start(timeout)

    logger.info("This is Paynt version {}.".format(version()))
    paynt.utils.version_check.check_stormpy_compatibility()

    # set CLI parameters
    paynt.quotient.quotient.Quotient.disable_expected_visits = disable_expected_visits
    paynt.synthesizer.synthesizer.Synthesizer.export_synthesis_filename_base = export_synthesis
    paynt.synthesizer.synthesizer_cegis.SynthesizerCEGIS.conflict_generator_type = ce_generator
    paynt.quotient.pomdp.PomdpQuotient.initial_memory_size = fsc_memory_size
    paynt.quotient.pomdp.PomdpQuotient.posterior_aware = posterior_aware
    paynt.quotient.decpomdp.DecPomdpQuotient.initial_memory_size = fsc_memory_size
    paynt.quotient.posmg.PosmgQuotient.initial_memory_size = fsc_memory_size

    paynt.quotient.mdp_family.MdpFamilyQuotient.initial_memory_size = fsc_memory_size

    paynt.synthesizer.policy_tree.SynthesizerPolicyTree.discard_unreachable_choices = mdp_discard_unreachable_choices

    paynt.synthesizer.decision_tree.SynthesizerDecisionTree.tree_depth = tree_depth
    paynt.synthesizer.decision_tree.SynthesizerDecisionTree.tree_enumeration = tree_enumeration
    paynt.synthesizer.decision_tree.SynthesizerDecisionTree.scheduler_path = tree_map_scheduler
    paynt.quotient.mdp.MdpQuotient.add_dont_care_action = add_dont_care_action

    storm_control = None
    if storm_pomdp:
        storm_control = paynt.quotient.storm_pomdp_control.StormPOMDPControl()
        storm_control.set_options(
            storm_options, get_storm_result, iterative_storm, use_storm_cutoffs,
            unfold_strategy_storm, prune_storm
        )

    sketch_path = os.path.join(project, sketch)
    properties_path = os.path.join(project, props)
    quotient = paynt.parser.sketch.Sketch.load_sketch(sketch_path, properties_path, export, relative_error, precision, constraint_bound, exact)
    synthesizer = paynt.synthesizer.synthesizer.Synthesizer.choose_synthesizer(quotient, method, fsc_synthesis, storm_control)
    synthesizer.run(optimum_threshold)

    if profiling:
        profiler.disable()
        print_profiler_stats(profiler)

def print_profiler_stats(profiler):
    stats = pstats.Stats(profiler)
    NUM_LINES = 10

    logger.debug("cProfiler info:")
    stats.sort_stats('tottime').print_stats(NUM_LINES)

    logger.debug("percentage breakdown:")
    entries = [ (key,data[2]) for key,data in stats.stats.items()]
    entries = sorted(entries, key=lambda x : x[1], reverse=True)
    entries = entries[:NUM_LINES]
    for key,data in entries:
        module,line,method = key
        if module == "~":
            callee = method
        else:
            callee = f"{module}:{line}({method})"
        percentage = round(data / stats.total_tt * 100,1)
        percentage = str(percentage).ljust(4)
        print(f"{percentage} %  {callee}")

def main():
    setup_logger()
    paynt_run()


if __name__ == "__main__":
    main()



import stormpy
from paynt.parser.drn_parser import DrnParser

class ModelBuilder:

    @classmethod
    def default_builder_options(cls, specification = None):
        # builder options
        if specification is not None:
            formulae = specification.stormpy_formulae()
            builder_options = stormpy.BuilderOptions(formulae)
        else:
            builder_options = stormpy.BuilderOptions()
        builder_options.set_build_state_valuations(True)
        builder_options.set_build_with_choice_origins(True)
        builder_options.set_build_all_labels(True)
        builder_options.set_build_choice_labels(True)
        builder_options.set_add_overlapping_guards_label(True)
        builder_options.set_build_observation_valuations(True)
        builder_options.set_build_all_reward_models(True)
        # builder_options.set_exploration_checks(True)
        return builder_options

    @classmethod
    def from_jani(cls, program, specification = None):
        builder_options = cls.default_builder_options(specification)
        builder_options.set_build_choice_labels(False)
        model = stormpy.build_sparse_model_with_options(program, builder_options)
        return model

    @classmethod
    def from_prism(cls, program, specification = None, use_exact=False):
        assert program.model_type in [stormpy.storage.PrismModelType.MDP, stormpy.storage.PrismModelType.POMDP]
        builder_options = cls.default_builder_options(specification)
        if use_exact:
            model = stormpy.build_sparse_exact_model_with_options(program, builder_options)
        else:
            model = stormpy.build_sparse_model_with_options(program, builder_options)
        return model

    @classmethod
    def from_drn(cls, drn_path, use_exact=False):
        return DrnParser.parse_drn(drn_path, use_exact)



import paynt.synthesizer.synthesizer
import paynt.synthesizer.synthesizer_ar
import paynt.synthesizer.synthesizer_cegis

import paynt.family.smt
import paynt.utils.timer

import logging
logger = logging.getLogger(__name__)


class StageControl:
    '''
    AR-CEGIS adaptivity: switch between ar/cegis, allocate more time to
    the more efficient method
    '''

    # whether only AR is performed
    only_ar = False
    # whether 1 AR followed by only CEGIS is performed
    only_cegis = False
    # whether adaptive hybrid is enabled
    adaptive_hybrid = True

    def __init__(self, family_size):
        # timings
        self.timer_ar = paynt.utils.timer.Timer()
        self.timer_cegis = paynt.utils.timer.Timer()

        self.family_size = family_size
        self.pruned_ar = 0
        self.pruned_cegis = 0
        
        # multiplier to derive time allocated for cegis
        # time_ar * factor = time_cegis
        # =1 is fair, >1 favours cegis, <1 favours ar
        self.cegis_efficiency = 1

    def start_ar(self):
        self.timer_cegis.stop()
        self.timer_ar.start()

    def start_cegis(self):
        self.timer_ar.stop()
        self.timer_cegis.start()

    def prune_ar(self, pruned):
        self.pruned_ar += pruned / self.family_size

    def prune_cegis(self, pruned):
        self.pruned_cegis += pruned / self.family_size

    def cegis_has_time(self):
        """
        :return True if cegis still has some time
        """
        
        # whether only AR is performed
        if StageControl.only_ar:
            return False

        # whether only CEGIS is performed
        if StageControl.only_cegis:
            return True

        # whether CEGIS has more time
        if self.timer_cegis.read() < self.timer_ar.read() * self.cegis_efficiency:
            return True

        # stop CEGIS
        self.timer_cegis.stop()

        if StageControl.adaptive_hybrid:
            if self.pruned_ar == 0 and self.pruned_cegis == 0:
                self.cegis_efficiency = 1
            elif self.pruned_ar == 0 and self.pruned_cegis > 0:
                self.cegis_efficiency = 2
            elif self.pruned_ar > 0 and self.pruned_cegis == 0:
                self.cegis_efficiency = 0.5
            else:
                success_rate_cegis = self.pruned_cegis / self.timer_cegis.read()
                success_rate_ar = self.pruned_ar / self.timer_ar.read()
                self.cegis_efficiency = success_rate_cegis / success_rate_ar
        
        return False


class SynthesizerHybrid(paynt.synthesizer.synthesizer_ar.SynthesizerAR, paynt.synthesizer.synthesizer_cegis.SynthesizerCEGIS):

    @property
    def method_name(self):
        return "hybrid"

    def synthesize_one(self, family):

        self.conflict_generator.initialize()
        smt_solver = paynt.family.smt.SmtSolver(self.quotient.family)

        # AR-CEGIS loop
        families = [family]
        self.stage_control = StageControl(family.size)
        while families:

            # initiate AR analysis
            self.stage_control.start_ar()
            
            # choose family
            family = families.pop(-1)

            # reset SMT solver level
            smt_solver.level(family.refinement_depth)

            # analyze the family
            self.verify_family(family)
            self.update_optimum(family)
            if family.analysis_result.can_improve == False:
                self.explore(family)
                self.stage_control.prune_ar(family.size)
                continue

            # undecided: initiate CEGIS analysis
            self.stage_control.start_cegis()

            # construct priority subfamily that corresponds to primary scheduler
            if family.analysis_result.optimality_result is not None:
                result = family.analysis_result.optimality_result
            else:
                result = family.analysis_result.constraints_result.results[0]
            priority_subfamily = family.assume_options_copy(result.primary_selection)

            # explore family assignments
            family_explored = False
            while True:

                if not self.stage_control.cegis_has_time():
                    break   # CEGIS timeout

                family.encode(smt_solver)
                # assignment = smt_solver.pick_assignment(family)
                assignment = smt_solver.pick_assignment_priority(family, priority_subfamily)
                if assignment is None:
                    family_explored = True
                    break   # explored whole family
                
                conflicts, accepting_assignment = self.analyze_family_assignment_cegis(family, assignment)
                pruned = smt_solver.exclude_conflicts(family, assignment, conflicts)
                self.explored += pruned
                self.stage_control.prune_cegis(pruned)

                if accepting_assignment is not None:
                    self.best_assignment = accepting_assignment
                    if not self.quotient.specification.can_be_improved:
                        return self.best_assignment

                # assignment is UNSAT: move on to the next assignment

            if family_explored:
                continue
        
            subfamilies = self.quotient.split(family)
            families = families + subfamilies

        return self.best_assignment
